# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2018
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import unittest

import keras.backend as k
import numpy as np

from art.attacks.evasion.elastic_net import ElasticNet
from art.estimators.estimator import BaseEstimator
from art.estimators.classification.classifier import ClassGradientsMixin
from art.estimators.classification.keras import KerasClassifier
from art.utils import random_targets, to_categorical
from tests.attacks.utils import backend_test_classifier_type_check_fail
from tests.utils import (
    TestBase,
    get_image_classifier_kr,
    get_image_classifier_pt,
    get_image_classifier_tf,
    get_tabular_classifier_kr,
    get_tabular_classifier_pt,
    get_tabular_classifier_tf,
    master_seed,
)

logger = logging.getLogger(__name__)


class TestElasticNet(TestBase):
    """
    A unittest class for testing the ElasticNet attack.
    """

    @classmethod
    def setUpClass(cls):
        master_seed(seed=1234)
        super().setUpClass()

        cls.n_train = 500
        cls.n_test = 10
        cls.x_train_mnist = cls.x_train_mnist[0 : cls.n_train]
        cls.y_train_mnist = cls.y_train_mnist[0 : cls.n_train]
        cls.x_test_mnist = cls.x_test_mnist[0 : cls.n_test]
        cls.y_test_mnist = cls.y_test_mnist[0 : cls.n_test]

    def setUp(self):
        master_seed(seed=1234)
        super().setUp()

    def test_2_tensorflow_failure_attack(self):
        """
        Test the corner case when attack fails.
        :return:
        """
        # Build TensorFlowClassifier
        tfc, sess = get_image_classifier_tf()

        # Failure attack
        ead = ElasticNet(
            classifier=tfc,
            targeted=True,
            max_iter=0,
            binary_search_steps=0,
            learning_rate=0,
            initial_const=1,
            verbose=False,
        )
        params = {"y": random_targets(self.y_test_mnist, tfc.nb_classes)}
        x_test_adv = ead.generate(self.x_test_mnist, **params)
        self.assertLessEqual(np.amax(x_test_adv), 1.0)
        self.assertGreaterEqual(np.amin(x_test_adv), 0.0)
        np.testing.assert_almost_equal(self.x_test_mnist, x_test_adv, 3)

        # Clean-up session
        if sess is not None:
            sess.close()

    def test_4_tensorflow_mnist(self):
        """
        First test with the TensorFlowClassifier.
        :return:
        """
        x_test_original = self.x_test_mnist.copy()

        # Build TensorFlowClassifier
        tfc, sess = get_image_classifier_tf(from_logits=True)

        # First attack
        ead = ElasticNet(classifier=tfc, targeted=True, max_iter=2, verbose=False)
        params = {"y": random_targets(self.y_test_mnist, tfc.nb_classes)}
        x_test_adv = ead.generate(self.x_test_mnist, **params)
        expected_x_test_adv = np.asarray(
            [
                0.45284095,
                0.43225235,
                0.577448,
                1.0,
                0.16554962,
                0.3587564,
                0.19202651,
                0.04293771,
                0.0,
                0.25811037,
                0.0,
                0.0290696,
                0.0,
                0.136172,
                0.6153389,
                0.12244645,
                0.0,
                0.7586619,
                0.8366919,
                0.22206311,
                0.12455986,
                0.02862802,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
            ]
        )
        np.testing.assert_array_almost_equal(x_test_adv[0, 14, :, 0], expected_x_test_adv, decimal=6)
        self.assertLessEqual(np.amax(x_test_adv), 1.0)
        self.assertGreaterEqual(np.amin(x_test_adv), 0.0)
        target = np.argmax(params["y"], axis=1)
        y_pred_adv = np.argmax(tfc.predict(x_test_adv), axis=1)
        logger.debug("EAD target: %s", target)
        logger.debug("EAD actual: %s", y_pred_adv)
        logger.info("EAD success rate on MNIST: %.2f%%", (100 * sum(target == y_pred_adv) / len(target)))
        self.assertTrue((target == y_pred_adv).any())

        # Second attack
        ead = ElasticNet(classifier=tfc, targeted=False, max_iter=2, verbose=False)
        params = {"y": random_targets(self.y_test_mnist, tfc.nb_classes)}
        x_test_adv = ead.generate(self.x_test_mnist, **params)
        self.assertLessEqual(np.amax(x_test_adv), 1.0)
        self.assertGreaterEqual(np.amin(x_test_adv), 0.0)
        target = np.argmax(params["y"], axis=1)
        y_pred_adv = np.argmax(tfc.predict(x_test_adv), axis=1)
        logger.debug("EAD target: %s", target)
        logger.debug("EAD actual: %s", y_pred_adv)
        logger.info("EAD success rate on MNIST: %.2f%%", (100 * sum(target != y_pred_adv) / float(len(target))))
        np.testing.assert_array_equal(y_pred_adv, np.asarray([7, 1, 1, 4, 4, 1, 4, 4, 4, 4]))

        # Third attack
        ead = ElasticNet(classifier=tfc, targeted=False, max_iter=2, verbose=False)
        params = {}
        x_test_adv = ead.generate(self.x_test_mnist, **params)
        expected_x_test_adv = np.asarray(
            [
                0.22766514,
                0.21726893,
                0.22802338,
                0.06168516,
                0.0,
                0.0,
                0.04722975,
                0.0,
                0.0,
                0.0,
                0.05455382,
                0.0,
                0.0,
                0.0,
                0.38886347,
                0.10553087,
                0.32285708,
                0.9794307,
                0.7589039,
                0.16586718,
                0.15969527,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
            ]
        )
        np.testing.assert_array_almost_equal(x_test_adv[0, 14, :, 0], expected_x_test_adv, decimal=6)
        self.assertLessEqual(np.amax(x_test_adv), 1.0)
        self.assertGreaterEqual(np.amin(x_test_adv), 0.0)
        y_pred = np.argmax(tfc.predict(self.x_test_mnist), axis=1)
        y_pred_adv = np.argmax(tfc.predict(x_test_adv), axis=1)
        logger.debug("EAD target: %s", y_pred)
        logger.debug("EAD actual: %s", y_pred_adv)
        logger.info("EAD success rate: %.2f%%", (100 * sum(y_pred != y_pred_adv) / float(len(y_pred))))
        np.testing.assert_array_equal(y_pred_adv, np.asarray([0, 4, 7, 9, 0, 7, 7, 3, 0, 7]))

        # First attack without batching
        ead_wob = ElasticNet(classifier=tfc, targeted=True, max_iter=2, batch_size=1, verbose=False)
        params = {"y": random_targets(self.y_test_mnist, tfc.nb_classes)}
        x_test_adv = ead_wob.generate(self.x_test_mnist, **params)
        expected_x_test_adv = np.asarray(
            [
                0.32568744,
                0.31085464,
                0.43235543,
                1.0,
                0.2433605,
                0.3411813,
                0.49469775,
                0.2603619,
                0.0,
                0.35584357,
                0.0,
                0.2046029,
                0.0,
                0.08249304,
                1.0,
                0.35813788,
                0.62945133,
                1.0,
                0.32154015,
                0.3497113,
                0.7613426,
                0.36533928,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
            ]
        )
        np.testing.assert_array_almost_equal(x_test_adv[0, 14, :, 0], expected_x_test_adv, decimal=6)
        self.assertLessEqual(np.amax(x_test_adv), 1.0)
        self.assertGreaterEqual(np.amin(x_test_adv), 0.0)
        target = np.argmax(params["y"], axis=1)
        y_pred_adv = np.argmax(tfc.predict(x_test_adv), axis=1)
        logger.debug("EAD target: %s", target)
        logger.debug("EAD actual: %s", y_pred_adv)
        logger.info("EAD success rate: %.2f%%", (100 * sum(target == y_pred_adv) / float(len(target))))
        self.assertTrue((target == y_pred_adv).any())

        # Second attack without batching
        ead_wob = ElasticNet(classifier=tfc, targeted=False, max_iter=2, batch_size=1, verbose=False)
        params = {"y": random_targets(self.y_test_mnist, tfc.nb_classes)}
        x_test_adv = ead_wob.generate(self.x_test_mnist, **params)
        self.assertLessEqual(np.amax(x_test_adv), 1.0)
        self.assertGreaterEqual(np.amin(x_test_adv), 0.0)
        target = np.argmax(params["y"], axis=1)
        y_pred_adv = np.argmax(tfc.predict(x_test_adv), axis=1)
        logger.debug("EAD target: %s", target)
        logger.debug("EAD actual: %s", y_pred_adv)
        logger.info("EAD success rate: %.2f%%", (100 * sum(target != y_pred_adv) / float(len(target))))
        np.testing.assert_array_equal(y_pred_adv, np.asarray([7, 1, 1, 4, 4, 1, 4, 4, 4, 4]))

        # Check that x_test has not been modified by attack and classifier
        self.assertAlmostEqual(float(np.max(np.abs(x_test_original - self.x_test_mnist))), 0.0, delta=0.00001)

        ead = ElasticNet(classifier=tfc, targeted=False, max_iter=2, verbose=False, decision_rule="L1")
        _ = ead.generate(self.x_test_mnist, **params)
        ead = ElasticNet(classifier=tfc, targeted=False, max_iter=2, verbose=False, decision_rule="L2")
        _ = ead.generate(self.x_test_mnist, **params)
        with self.assertRaises(ValueError):
            _ = ElasticNet(classifier=tfc, targeted=False, max_iter=2, verbose=False, decision_rule="L3")

        # Close session
        if sess is not None:
            sess.close()

    def test_9a_keras_mnist(self):
        """
        Second test with the KerasClassifier.
        :return:
        """
        x_test_original = self.x_test_mnist.copy()

        # Build KerasClassifier
        krc = get_image_classifier_kr()

        # First attack
        ead = ElasticNet(classifier=krc, targeted=True, max_iter=2, verbose=False)
        y_target = to_categorical(np.asarray([6, 6, 7, 4, 9, 7, 9, 0, 1, 0]), nb_classes=10)
        x_test_adv = ead.generate(self.x_test_mnist, y=y_target)
        expected_x_test_adv = np.asarray(
            [
                0.0000000e00,
                0.0000000e00,
                0.0000000e00,
                0.0000000e00,
                0.0000000e00,
                0.0000000e00,
                0.0000000e00,
                0.0000000e00,
                0.0000000e00,
                7.8193319e-04,
                5.5843666e-03,
                0.0000000e00,
                0.0000000e00,
                4.9869284e-01,
                1.0000000e00,
                6.4663666e-01,
                3.4855194e-03,
                3.5087438e-03,
                0.0000000e00,
                9.8862723e-03,
                3.8835173e-03,
                3.0151173e-02,
                0.0000000e00,
                0.0000000e00,
                0.0000000e00,
                0.0000000e00,
                0.0000000e00,
                0.0000000e00,
            ]
        )
        np.testing.assert_array_almost_equal(x_test_adv[2, 14, :, 0], expected_x_test_adv, decimal=6)
        self.assertLessEqual(np.amax(x_test_adv), 1.0)
        self.assertGreaterEqual(np.amin(x_test_adv), 0.0)
        target = np.argmax(y_target, axis=1)
        y_pred_adv = np.argmax(krc.predict(x_test_adv), axis=1)
        logger.debug("EAD target: %s", target)
        logger.debug("EAD actual: %s", y_pred_adv)
        logger.info("EAD success rate: %.2f%%", (100 * sum(target == y_pred_adv) / float(len(target))))
        self.assertTrue((target == y_pred_adv).any())

        # Second attack
        ead = ElasticNet(classifier=krc, targeted=False, max_iter=2, verbose=False)
        y_target = to_categorical(np.asarray([9, 5, 6, 7, 1, 6, 1, 5, 8, 5]), nb_classes=10)
        x_test_adv = ead.generate(self.x_test_mnist, y=y_target)
        self.assertLessEqual(np.amax(x_test_adv), 1.0)
        self.assertGreaterEqual(np.amin(x_test_adv), 0.0)
        y_pred_adv = np.argmax(krc.predict(x_test_adv), axis=1)
        logger.debug("EAD target: %s", y_target)
        logger.debug("EAD actual: %s", y_pred_adv)
        logger.info("EAD success rate: %.2f", (100 * sum(target != y_pred_adv) / float(len(target))))
        self.assertTrue((target != y_pred_adv).any())
        np.testing.assert_array_equal(y_pred_adv, np.asarray([7, 1, 1, 4, 4, 1, 4, 4, 4, 4]))

        # Check that x_test has not been modified by attack and classifier
        self.assertAlmostEqual(float(np.max(np.abs(x_test_original - self.x_test_mnist))), 0.0, delta=0.00001)

        k.clear_session()

    def test_6_pytorch_mnist(self):
        """
        Third test with the PyTorchClassifier.
        :return:
        """
        x_test = np.reshape(self.x_test_mnist, (self.x_test_mnist.shape[0], 1, 28, 28)).astype(np.float32)
        x_test_original = x_test.copy()

        # Build PyTorchClassifier
        ptc = get_image_classifier_pt(from_logits=False)

        # First attack
        ead = ElasticNet(classifier=ptc, targeted=True, max_iter=2, verbose=False)
        params = {"y": random_targets(self.y_test_mnist, ptc.nb_classes)}
        x_test_adv = ead.generate(x_test, **params)
        expected_x_test_adv = np.asarray(
            [
                0.01758931,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.00698278,
                0.0,
                0.11318438,
                0.36223832,
                0.54720753,
                0.93125045,
                1.0,
                0.9999359,
                0.8638486,
                0.6354147,
                0.5600332,
                0.24081531,
                0.25882354,
                0.00899846,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
            ]
        )
        np.testing.assert_array_almost_equal(x_test_adv[2, 0, :, 14], expected_x_test_adv, decimal=6)
        self.assertLessEqual(np.amax(x_test_adv), 1.0)
        self.assertGreaterEqual(np.amin(x_test_adv), 0.0)
        target = np.argmax(params["y"], axis=1)
        y_pred_adv = np.argmax(ptc.predict(x_test_adv), axis=1)
        self.assertTrue((target == y_pred_adv).any())

        # Second attack
        ead = ElasticNet(classifier=ptc, targeted=False, max_iter=2, verbose=False)
        params = {"y": random_targets(self.y_test_mnist, ptc.nb_classes)}
        x_test_adv = ead.generate(x_test, **params)
        self.assertLessEqual(np.amax(x_test_adv), 1.0)
        self.assertGreaterEqual(np.amin(x_test_adv), 0.0)
        target = np.argmax(params["y"], axis=1)
        y_pred_adv = np.argmax(ptc.predict(x_test_adv), axis=1)
        self.assertTrue((target != y_pred_adv).any())
        np.testing.assert_array_equal(y_pred_adv, np.asarray([7, 1, 1, 4, 4, 1, 4, 4, 4, 4]))

        # Check that x_test has not been modified by attack and classifier
        self.assertAlmostEqual(float(np.max(np.abs(x_test_original - x_test))), 0.0, delta=0.00001)

    def test_1_classifier_type_check_fail(self):
        backend_test_classifier_type_check_fail(ElasticNet, [BaseEstimator, ClassGradientsMixin])

    def test_8_keras_iris_clipped(self):
        classifier = get_tabular_classifier_kr()
        attack = ElasticNet(classifier, targeted=False, max_iter=10, verbose=False)
        x_test_adv = attack.generate(self.x_test_iris)
        expected_x_test_adv = np.asarray([0.8670352, 0.4624909, 0.6453267, 0.23096858])
        np.testing.assert_array_almost_equal(x_test_adv[0, :], expected_x_test_adv, decimal=6)
        self.assertLessEqual(np.amax(x_test_adv), 1.0)
        self.assertGreaterEqual(np.amin(x_test_adv), 0.0)
        predictions_adv = np.argmax(classifier.predict(x_test_adv), axis=1)
        np.testing.assert_array_equal(
            predictions_adv,
            np.asarray(
                [
                    1,
                    1,
                    1,
                    2,
                    1,
                    1,
                    1,
                    2,
                    1,
                    2,
                    1,
                    1,
                    1,
                    2,
                    1,
                    1,
                    2,
                    2,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    2,
                    1,
                    2,
                    1,
                    2,
                    1,
                    0,
                    1,
                    1,
                    1,
                    2,
                    2,
                    2,
                    2,
                    1,
                    1,
                    2,
                ]
            ),
        )
        accuracy = 1.0 - np.sum(predictions_adv == np.argmax(self.y_test_iris, axis=1)) / self.y_test_iris.shape[0]
        logger.info("EAD success rate on Iris: %.2f%%", (accuracy * 100))

        k.clear_session()

    def test_9_keras_iris_unbounded(self):
        classifier = get_tabular_classifier_kr()

        # Recreate a classifier without clip values
        classifier = KerasClassifier(model=classifier._model, use_logits=False, channels_first=True)
        attack = ElasticNet(classifier, targeted=False, max_iter=10, verbose=False)
        x_test_adv = attack.generate(self.x_test_iris)
        expected_x_test_adv = np.asarray([0.8670352, 0.4624909, 0.6453267, 0.23096858])
        np.testing.assert_array_almost_equal(x_test_adv[0, :], expected_x_test_adv, decimal=6)
        predictions_adv = np.argmax(classifier.predict(x_test_adv), axis=1)
        np.testing.assert_array_equal(
            predictions_adv,
            np.asarray(
                [
                    1,
                    1,
                    1,
                    2,
                    1,
                    1,
                    1,
                    2,
                    1,
                    2,
                    1,
                    1,
                    1,
                    2,
                    1,
                    1,
                    2,
                    2,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    2,
                    1,
                    2,
                    1,
                    2,
                    1,
                    0,
                    1,
                    1,
                    1,
                    2,
                    2,
                    2,
                    2,
                    1,
                    1,
                    2,
                ]
            ),
        )
        accuracy = 1.0 - np.sum(predictions_adv == np.argmax(self.y_test_iris, axis=1)) / self.y_test_iris.shape[0]
        logger.info("EAD success rate on Iris: %.2f%%", (accuracy * 100))

        k.clear_session()

    def test_3_tensorflow_iris(self):
        classifier, sess = get_tabular_classifier_tf()

        # Test untargeted attack
        attack = ElasticNet(classifier, targeted=False, max_iter=10, verbose=False)
        x_test_adv = attack.generate(self.x_test_iris)
        expected_x_test_adv = np.asarray([0.84810126, 0.43320203, 0.70404345, 0.29160658])
        np.testing.assert_array_almost_equal(x_test_adv[0, :], expected_x_test_adv, decimal=6)
        self.assertLessEqual(np.amax(x_test_adv), 1.0)
        self.assertGreaterEqual(np.amin(x_test_adv), 0.0)

        predictions_adv = np.argmax(classifier.predict(x_test_adv), axis=1)
        np.testing.assert_array_equal(
            predictions_adv,
            np.asarray(
                [
                    1,
                    2,
                    2,
                    2,
                    1,
                    1,
                    1,
                    2,
                    1,
                    2,
                    1,
                    1,
                    1,
                    2,
                    2,
                    2,
                    2,
                    2,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    2,
                    2,
                    2,
                    2,
                    1,
                    2,
                    1,
                    2,
                    1,
                    0,
                    1,
                    2,
                    1,
                    2,
                    0,
                    2,
                    2,
                    1,
                    1,
                    2,
                ]
            ),
        )
        accuracy = 1.0 - np.sum(predictions_adv == np.argmax(self.y_test_iris, axis=1)) / self.y_test_iris.shape[0]
        logger.info("EAD success rate on Iris: %.2f%%", (accuracy * 100))

        # Test targeted attack
        targets = random_targets(self.y_test_iris, nb_classes=3)
        attack = ElasticNet(classifier, targeted=True, max_iter=10, verbose=False)
        x_test_adv = attack.generate(self.x_test_iris, **{"y": targets})
        expected_x_test_adv = np.asarray([0.88713187, 0.5239736, 0.49900988, 0.05677444])
        np.testing.assert_array_almost_equal(x_test_adv[0, :], expected_x_test_adv, decimal=6)
        self.assertLessEqual(np.amax(x_test_adv), 1.0)
        self.assertGreaterEqual(np.amin(x_test_adv), 0.0)

        predictions_adv = np.argmax(classifier.predict(x_test_adv), axis=1)
        np.testing.assert_array_equal(
            predictions_adv,
            np.asarray(
                [
                    0,
                    0,
                    0,
                    2,
                    2,
                    2,
                    2,
                    2,
                    2,
                    2,
                    2,
                    2,
                    0,
                    2,
                    0,
                    0,
                    2,
                    2,
                    0,
                    2,
                    2,
                    2,
                    2,
                    2,
                    2,
                    0,
                    0,
                    0,
                    2,
                    0,
                    2,
                    2,
                    2,
                    2,
                    2,
                    0,
                    0,
                    0,
                    2,
                    2,
                    2,
                    2,
                    2,
                    0,
                    2,
                ]
            ),
        )

        accuracy = np.sum(predictions_adv == np.argmax(targets, axis=1)) / self.y_test_iris.shape[0]
        logger.info("Targeted EAD success rate on Iris: %.2f%%", (accuracy * 100))

        # Close session
        if sess is not None:
            sess.close()

    def test_5_pytorch_iris(self):
        classifier = get_tabular_classifier_pt()
        attack = ElasticNet(classifier, targeted=False, max_iter=10, verbose=False)
        x_test_adv = attack.generate(self.x_test_iris.astype(np.float32))
        expected_x_test_adv = np.asarray([0.84810126, 0.43320203, 0.70404345, 0.29160658])
        np.testing.assert_array_almost_equal(x_test_adv[0, :], expected_x_test_adv, decimal=6)
        self.assertLessEqual(np.amax(x_test_adv), 1.0)
        self.assertGreaterEqual(np.amin(x_test_adv), 0.0)

        predictions_adv = np.argmax(classifier.predict(x_test_adv), axis=1)
        np.testing.assert_array_equal(
            predictions_adv,
            np.asarray(
                [
                    1,
                    2,
                    2,
                    2,
                    1,
                    1,
                    1,
                    2,
                    1,
                    2,
                    1,
                    1,
                    1,
                    2,
                    2,
                    2,
                    2,
                    2,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    2,
                    2,
                    2,
                    2,
                    1,
                    2,
                    1,
                    2,
                    1,
                    0,
                    1,
                    2,
                    1,
                    2,
                    0,
                    2,
                    2,
                    1,
                    1,
                    2,
                ]
            ),
        )

        accuracy = 1.0 - np.sum(predictions_adv == np.argmax(self.y_test_iris, axis=1)) / self.y_test_iris.shape[0]
        logger.info("EAD success rate on Iris: %.2f%%", (accuracy * 100))

    def test_7_scikitlearn(self):
        from sklearn.linear_model import LogisticRegression
        from sklearn.svm import SVC, LinearSVC

        from art.estimators.classification.scikitlearn import SklearnClassifier

        scikitlearn_test_cases = [
            LogisticRegression(solver="lbfgs", multi_class="auto"),
            SVC(gamma="auto"),
            LinearSVC(),
        ]

        x_test_original = self.x_test_iris.copy()

        for model in scikitlearn_test_cases:
            classifier = SklearnClassifier(model=model, clip_values=(0, 1))
            classifier.fit(x=self.x_test_iris, y=self.y_test_iris)

            # Test untargeted attack
            attack = ElasticNet(classifier, targeted=False, max_iter=2, verbose=False)
            x_test_adv = attack.generate(self.x_test_iris)
            self.assertFalse((self.x_test_iris == x_test_adv).all())
            self.assertLessEqual(np.amax(x_test_adv), 1.0)
            self.assertGreaterEqual(np.amin(x_test_adv), 0.0)

            predictions_adv = np.argmax(classifier.predict(x_test_adv), axis=1)
            self.assertFalse((np.argmax(self.y_test_iris, axis=1) == predictions_adv).all())
            accuracy = 1.0 - np.sum(predictions_adv == np.argmax(self.y_test_iris, axis=1)) / self.y_test_iris.shape[0]
            logger.info("EAD success rate  of " + classifier.__class__.__name__ + " on Iris: %.2f%%", (accuracy * 100))

            # Test targeted attack
            targets = random_targets(self.y_test_iris, nb_classes=3)
            attack = ElasticNet(classifier, targeted=True, max_iter=2, verbose=False)
            x_test_adv = attack.generate(self.x_test_iris, **{"y": targets})
            self.assertFalse((self.x_test_iris == x_test_adv).all())
            self.assertLessEqual(np.amax(x_test_adv), 1.0)
            self.assertGreaterEqual(np.amin(x_test_adv), 0.0)

            predictions_adv = np.argmax(classifier.predict(x_test_adv), axis=1)
            self.assertTrue((np.argmax(targets, axis=1) == predictions_adv).any())
            accuracy = np.sum(predictions_adv == np.argmax(targets, axis=1)) / self.y_test_iris.shape[0]
            logger.info(
                "Targeted EAD success rate of " + classifier.__class__.__name__ + " on Iris: %.2f%%", (accuracy * 100)
            )

            # Check that x_test has not been modified by attack and classifier
            self.assertAlmostEqual(float(np.max(np.abs(x_test_original - self.x_test_iris))), 0.0, delta=0.00001)

    def test_check_params(self):

        ptc = get_image_classifier_pt(from_logits=True)

        with self.assertRaises(ValueError):
            _ = ElasticNet(ptc, binary_search_steps="1.0")
        with self.assertRaises(ValueError):
            _ = ElasticNet(ptc, binary_search_steps=-1)

        with self.assertRaises(ValueError):
            _ = ElasticNet(ptc, max_iter="1.0")
        with self.assertRaises(ValueError):
            _ = ElasticNet(ptc, max_iter=-1)

        with self.assertRaises(ValueError):
            _ = ElasticNet(ptc, batch_size="1.0")
        with self.assertRaises(ValueError):
            _ = ElasticNet(ptc, batch_size=-1)

        with self.assertRaises(ValueError):
            _ = ElasticNet(ptc, decision_rule=1.0)

        with self.assertRaises(ValueError):
            _ = ElasticNet(ptc, verbose="True")


if __name__ == "__main__":
    unittest.main()



"""
This module contains certified classifiers.
"""
import importlib
from art.estimators.certification.randomized_smoothing.randomized_smoothing import RandomizedSmoothingMixin
from art.estimators.certification.randomized_smoothing.numpy import NumpyRandomizedSmoothing
from art.estimators.certification.randomized_smoothing.tensorflow import TensorFlowV2RandomizedSmoothing
from art.estimators.certification.randomized_smoothing.pytorch import PyTorchRandomizedSmoothing
from art.estimators.certification.derandomized_smoothing.derandomized_smoothing import DeRandomizedSmoothingMixin
from art.estimators.certification.derandomized_smoothing.pytorch import PyTorchDeRandomizedSmoothing
from art.estimators.certification.derandomized_smoothing.tensorflow import TensorFlowV2DeRandomizedSmoothing

if importlib.util.find_spec("torch") is not None:
    from art.estimators.certification.deep_z.deep_z import ZonoDenseLayer
    from art.estimators.certification.deep_z.deep_z import ZonoBounds
    from art.estimators.certification.deep_z.deep_z import ZonoConv
    from art.estimators.certification.deep_z.deep_z import ZonoReLU
    from art.estimators.certification.deep_z.pytorch import PytorchDeepZ
    from art.estimators.certification.interval.interval import PyTorchIntervalDense
    from art.estimators.certification.interval.interval import PyTorchIntervalConv2D
    from art.estimators.certification.interval.interval import PyTorchIntervalReLU
    from art.estimators.certification.interval.interval import PyTorchIntervalFlatten
    from art.estimators.certification.interval.interval import PyTorchIntervalBounds
    from art.estimators.certification.interval.pytorch import PyTorchIBPClassifier
else:
    import warnings

    warnings.warn("PyTorch not found. Not importing DeepZ or Interval Bound Propagation functionality")



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2022
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import numpy as np
import pytest
import os

from art.attacks.poisoning.perturbations.audio_perturbations import CacheToneTrigger, CacheAudioTrigger

from tests.utils import ARTTestException

logger = logging.getLogger(__name__)


@pytest.mark.framework_agnostic
def test_insert_tone_trigger(art_warning):
    try:
        # test single example
        trigger = CacheToneTrigger(sampling_rate=16000)
        audio = trigger.insert(x=np.zeros(3200))
        assert audio.shape == (3200,)
        assert np.max(audio) != 0
        assert np.max(np.abs(audio)) <= 1.0

        # test single example with differet duration, frequency, and scale
        trigger = CacheToneTrigger(sampling_rate=16000, frequency=16000, duration=0.2, scale=0.5)
        audio = trigger.insert(x=np.zeros(3200))
        assert audio.shape == (3200,)
        assert np.max(audio) != 0

        # test a batch of examples
        audio = trigger.insert(x=np.zeros((10, 3200)))
        assert audio.shape == (10, 3200)
        assert np.max(audio) != 0

        # test single example with shift
        trigger = CacheToneTrigger(sampling_rate=16000, shift=10)
        audio = trigger.insert(x=np.zeros(3200))
        assert audio.shape == (3200,)
        assert np.max(audio) != 0
        assert np.sum(audio[:10]) == 0

        # test a batch of examples with random shift
        trigger = CacheToneTrigger(sampling_rate=16000, random=True)
        audio = trigger.insert(x=np.zeros((10, 3200)))
        assert audio.shape == (10, 3200)
        assert np.max(audio) != 0

        # test when length of backdoor is larger than that of audio signal
        with pytest.raises(ValueError):
            trigger = CacheToneTrigger(sampling_rate=16000, duration=0.3)
            _ = trigger.insert(x=np.zeros(3200))

        # test when shift + backdoor is larger than that of audio signal
        with pytest.raises(ValueError):
            trigger = CacheToneTrigger(sampling_rate=16000, duration=0.2, shift=5)
            _ = trigger.insert(x=np.zeros(3200))

    except ARTTestException as e:
        art_warning(e)


@pytest.mark.framework_agnostic
def test_insert_audio_trigger(art_warning):
    file_path = os.path.join(os.getcwd(), "utils/data/backdoors/cough_trigger.wav")
    try:
        # test single example
        trigger = CacheAudioTrigger(sampling_rate=16000, backdoor_path=file_path)
        audio = trigger.insert(x=np.zeros(32000))
        assert audio.shape == (32000,)
        assert np.max(audio) != 0
        assert np.max(np.abs(audio)) <= 1.0

        # test single example with differet duration and scale
        trigger = CacheAudioTrigger(
            sampling_rate=16000,
            backdoor_path=file_path,
            duration=0.8,
            scale=0.5,
        )
        audio = trigger.insert(x=np.zeros(32000))
        assert audio.shape == (32000,)
        assert np.max(audio) != 0

        # test a batch of examples
        trigger = CacheAudioTrigger(sampling_rate=16000, backdoor_path=file_path)
        audio = trigger.insert(x=np.zeros((10, 16000)))

        assert audio.shape == (10, 16000)
        assert np.max(audio) != 0

        # test single example with shift
        trigger = CacheAudioTrigger(sampling_rate=16000, backdoor_path=file_path, shift=10)
        audio = trigger.insert(x=np.zeros(32000))
        assert audio.shape == (32000,)
        assert np.max(audio) != 0
        assert np.sum(audio[:10]) == 0

        # test a batch of examples with random shift
        trigger = CacheAudioTrigger(sampling_rate=16000, backdoor_path=file_path, random=True)
        audio = trigger.insert(x=np.zeros((10, 32000)))
        assert audio.shape == (10, 32000)
        assert np.max(audio) != 0

        # test when length of backdoor is larger than that of audio signal
        with pytest.raises(ValueError):
            trigger = CacheAudioTrigger(sampling_rate=16000, backdoor_path=file_path)
            _ = trigger.insert(x=np.zeros(15000))

        # test when shift + backdoor is larger than that of audio signal
        with pytest.raises(ValueError):
            trigger = CacheAudioTrigger(
                sampling_rate=16000,
                backdoor_path=file_path,
                duration=1,
                shift=5,
            )
            _ = trigger.insert(x=np.zeros(16000))

    except ARTTestException as e:
        art_warning(e)



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2023
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module implements adversarial training with TRADES protocol.

| Paper link: https://proceedings.mlr.press/v97/zhang19p.html

| It was noted that this protocol uses a modified loss called TRADES loss which is a combination of cross entropy
loss on clean data and KL divergence loss between clean data and adversarial data. Consequently, framework specific
implementations are being provided in ART.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import abc
from typing import Optional, Tuple, TYPE_CHECKING

import numpy as np

from art.defences.trainer.trainer import Trainer
from art.attacks.attack import EvasionAttack
from art.data_generators import DataGenerator

if TYPE_CHECKING:
    from art.utils import CLASSIFIER_LOSS_GRADIENTS_TYPE


class AdversarialTrainerTRADES(Trainer, abc.ABC):
    """
    This is abstract class for different backend-specific implementations of TRADES protocol
    for adversarial training.

    | Paper link: https://proceedings.mlr.press/v97/zhang19p.html
    """

    def __init__(
        self,
        classifier: "CLASSIFIER_LOSS_GRADIENTS_TYPE",
        attack: EvasionAttack,
        beta: float = 6.0,
    ):
        """
        Create an :class:`.AdversarialTrainerTRADES` instance.

        :param classifier: Model to train adversarially.
        :param attack: attack to use for data augmentation in adversarial training
        :param beta: The scaling factor controlling tradeoff between clean loss and adversarial loss
        """
        self._attack = attack
        self._beta = beta
        super().__init__(classifier)

    @abc.abstractmethod
    def fit(  # pylint: disable=W0221
        self,
        x: np.ndarray,
        y: np.ndarray,
        validation_data: Optional[Tuple[np.ndarray, np.ndarray]] = None,
        batch_size: int = 128,
        nb_epochs: int = 20,
        **kwargs
    ):
        """
        Train a model adversarially with TRADES. See class documentation for more information on the exact procedure.

        :param x: Training set.
        :param y: Labels for the training set.
        :param validation_data: Tuple consisting of validation data, (x_val, y_val)
        :param batch_size: Size of batches.
        :param nb_epochs: Number of epochs to use for trainings.
        :param kwargs: Dictionary of framework-specific arguments. These will be passed as such to the `fit` function of
               the target classifier.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def fit_generator(self, generator: DataGenerator, nb_epochs: int = 20, **kwargs):
        """
        Train a model adversarially using a data generator.
        See class documentation for more information on the exact procedure.

        :param generator: Data generator.
        :param nb_epochs: Number of epochs to use for trainings.
        :param kwargs: Dictionary of framework-specific arguments. These will be passed as such to the `fit` function of
               the target classifier.
        """
        raise NotImplementedError

    def predict(self, x: np.ndarray, **kwargs) -> np.ndarray:
        """
        Perform prediction using the adversarially trained classifier.

        :param x: Input samples.
        :param kwargs: Other parameters to be passed on to the `predict` function of the classifier.
        :return: Predictions for test set.
        """
        return self._classifier.predict(x, **kwargs)



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2022
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module implements Expectation over Transformation preprocessing for image rotation in PyTorch.
"""
import logging
from typing import Dict, List, Optional, TYPE_CHECKING, Tuple, Union

import numpy as np

from art.preprocessing.expectation_over_transformation.pytorch import EoTPyTorch

if TYPE_CHECKING:
    # pylint: disable=C0412
    import torch

logger = logging.getLogger(__name__)


class EoTImageRotationPyTorch(EoTPyTorch):
    """
    This module implements Expectation over Transformation preprocessing for image rotation in PyTorch.
    """

    params = ["nb_samples", "angles", "clip_values", "label_type"]

    label_types = ["classification", "object_detection"]

    def __init__(
        self,
        nb_samples: int,
        clip_values: Tuple[float, float],
        angles: Union[float, Tuple[float, float], List[float]] = 45.0,
        label_type: str = "classification",
        apply_fit: bool = False,
        apply_predict: bool = True,
    ) -> None:
        """
        Create an instance of EoTImageRotationPyTorch.

        :param nb_samples: Number of random samples per input sample.
        :param clip_values: Tuple of the form `(min, max)` representing the minimum and maximum values allowed
                            for features.
        :param angles: In degrees and counter-clockwise. If a positive scalar it defines the uniform sampling range from
                       negative to positive value. If a tuple of two scalar angles it defines the uniform sampling
                       range from minimum to maximum angles. If a list of scalar values it defines the discrete angles
                       that will be sampled. For `label_type="object_detection"` only a list of multiples of 90 degrees
                       is supported.
        :param label_type: String defining the type of labels. Currently supported: `classification`, `object_detection`
        :param apply_fit: True if applied during fitting/training.
        :param apply_predict: True if applied during predicting.
        """
        super().__init__(
            apply_fit=apply_fit, apply_predict=apply_predict, nb_samples=nb_samples, clip_values=clip_values
        )

        self.angles = angles
        self.angles_range = (-angles, angles) if isinstance(angles, (int, float)) else angles
        self.label_type = label_type
        self._check_params()

    def _transform(
        self, x: "torch.Tensor", y: Optional[Union["torch.Tensor", List[Dict[str, "torch.Tensor"]]]], **kwargs
    ) -> Tuple["torch.Tensor", Optional[Union["torch.Tensor", List[Dict[str, "torch.Tensor"]]]]]:
        """
        Transformation of an input image and its label by randomly sampled rotation.

        :param x: Input samples.
        :param y: Label of the samples `x`.
        :return: Transformed samples and labels.
        """
        import torch
        import torchvision

        if isinstance(self.angles, list):
            angles = np.random.choice(self.angles).item()
        else:
            angles = np.random.uniform(low=self.angles_range[0], high=self.angles_range[1])

        # Ensure channels-first
        channels_first = True
        if x.shape[-1] in [1, 3]:
            x = torch.permute(x, (0, 3, 1, 2))
            channels_first = False

        expand = self.label_type == "object_detection"

        x_preprocess = torchvision.transforms.functional.rotate(
            img=x,
            angle=angles,
            interpolation=torchvision.transforms.functional.InterpolationMode.NEAREST,
            expand=expand,
        )

        x_preprocess = torch.clamp(
            input=x_preprocess,
            min=-self.clip_values[0],
            max=self.clip_values[1],
        )

        y_preprocess: Optional[Union["torch.Tensor", List[Dict[str, "torch.Tensor"]]]]

        if self.label_type == "object_detection" and y is not None:

            y_od: List[Dict[str, "torch.Tensor"]] = [{}]

            if isinstance(y, list):
                if isinstance(y[0], dict):
                    y_od[0]["boxes"] = torch.clone(y[0]["boxes"])
                    y_od[0]["labels"] = torch.clone(y[0]["labels"])
                else:
                    raise TypeError("Wrong type for `y` and label_type=object_detection.")
            else:
                raise TypeError("Wrong type for `y` and label_type=object_detection.")

            y_b = y[0]["boxes"]
            image_width = x.shape[2]
            image_height = x.shape[1]
            x_1_arr = y_b[:, 0]
            y_1_arr = y_b[:, 1]
            x_2_arr = y_b[:, 2]
            y_2_arr = y_b[:, 3]
            box_width = x_2_arr - x_1_arr
            box_height = y_2_arr - y_1_arr

            if angles == 0:
                x_1_new = x_1_arr
                y_1_new = y_1_arr
                x_2_new = x_2_arr
                y_2_new = y_2_arr

            elif angles == 90:
                x_1_new = y_1_arr
                y_1_new = image_width - x_1_arr - box_width
                x_2_new = y_1_arr + box_height
                y_2_new = image_width - x_1_arr

            elif angles == 180:
                x_1_new = image_width - x_2_arr
                y_1_new = image_height - y_2_arr
                x_2_new = x_1_new + box_width
                y_2_new = y_1_new + box_height

            elif angles == 270:
                x_1_new = image_height - y_1_arr - box_height
                y_1_new = x_1_arr
                x_2_new = image_height - y_1_arr
                y_2_new = x_1_arr + box_width

            else:
                raise ValueError("The angle is not supported for object detection.")

            y_od[0]["boxes"][:, 0] = x_1_new
            y_od[0]["boxes"][:, 1] = y_1_new
            y_od[0]["boxes"][:, 2] = x_2_new
            y_od[0]["boxes"][:, 3] = y_2_new

            y_preprocess = y_od

        else:

            y_preprocess = y

        if not channels_first:
            x_preprocess = torch.permute(x_preprocess, (0, 2, 3, 1))

        return x_preprocess, y_preprocess

    def _check_params(self) -> None:

        # pylint: disable=R0916
        if (
            self.label_type == "classification"
            and not isinstance(self.angles, (int, float, tuple, list))
            or (
                isinstance(self.angles, tuple)
                and (
                    len(self.angles) != 2
                    or not isinstance(self.angles[0], (int, float))
                    or not isinstance(self.angles[1], (int, float))
                    or self.angles[0] > self.angles[1]
                )
            )
        ):
            raise ValueError("The range of angles must be a float in the range (0.0, 180.0].")

        if self.label_type not in self.label_types:
            raise ValueError(
                f"The input for label_type needs to be one of {self.label_types},"
                f"currently receiving `{self.label_type}`."
            )

        if self.label_type == "object_detection":
            if not isinstance(self.angles, list):
                raise ValueError(
                    """For `label_type="object_detection"` only a list of multiples of 90 degrees is supported."""
                )
            if isinstance(self.angles, list):
                for angle in self.angles:
                    if divmod(angle, 90)[1] != 0:
                        raise ValueError(
                            """For `label_type="object_detection"` only a list of multiples of 90 degrees is
                               supported."""
                        )



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2023
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module implements the BadDet Object Disappearance Attack (ODA) on object detectors.

| Paper link: https://arxiv.org/abs/2205.14497
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
from typing import Dict, List, Tuple, Union

import numpy as np
from tqdm.auto import tqdm

from art.attacks.attack import PoisoningAttackObjectDetector
from art.attacks.poisoning.backdoor_attack import PoisoningAttackBackdoor

logger = logging.getLogger(__name__)


class BadDetObjectDisappearanceAttack(PoisoningAttackObjectDetector):
    """
    Implementation of the BadDet Object Disappearance Attack.

    | Paper link: https://arxiv.org/abs/2205.14497
    """

    attack_params = PoisoningAttackObjectDetector.attack_params + [
        "backdoor",
        "class_source",
        "percent_poison",
        "channels_first",
        "verbose",
    ]
    _estimator_requirements = ()

    def __init__(
        self,
        backdoor: PoisoningAttackBackdoor,
        class_source: int = 0,
        percent_poison: float = 0.3,
        channels_first: bool = False,
        verbose: bool = False,
    ) -> None:
        """
        Creates a new BadDet Object Disappearance Attack

        :param backdoor: the backdoor chosen for this attack.
        :param class_source: The source class from which triggers were selected.
        :param percent_poison: The ratio of samples to poison in the source class, with range [0, 1].
        :param channels_first: Set channels first or last.
        :param verbose: Show progress bars.
        """
        super().__init__()
        self.backdoor = backdoor
        self.class_source = class_source
        self.percent_poison = percent_poison
        self.channels_first = channels_first
        self.verbose = verbose
        self._check_params()

    def poison(  # pylint: disable=W0221
        self,
        x: Union[np.ndarray, List[np.ndarray]],
        y: List[Dict[str, np.ndarray]],
        **kwargs,
    ) -> Tuple[Union[np.ndarray, List[np.ndarray]], List[Dict[str, np.ndarray]]]:
        """
        Generate poisoning examples by inserting the backdoor onto the input `x` and changing the classification
        for labels `y`.

        :param x: Sample images of shape `NCHW` or `NHWC` or a list of sample images of any size.
        :param y: True labels of type `List[Dict[np.ndarray]]`, one dictionary per input image. The keys and values
                  of the dictionary are:

                  - boxes [N, 4]: the boxes in [x1, y1, x2, y2] format, with 0 <= x1 < x2 <= W and 0 <= y1 < y2 <= H.
                  - labels [N]: the labels for each image.
        :return: An tuple holding the `(poisoning_examples, poisoning_labels)`.
        """
        if isinstance(x, np.ndarray):
            x_ndim = len(x.shape)
        else:
            x_ndim = len(x[0].shape) + 1

        if x_ndim != 4:
            raise ValueError("Unrecognized input dimension. BadDet ODA can only be applied to image data.")

        # copy images
        x_poison: Union[np.ndarray, List[np.ndarray]]
        if isinstance(x, np.ndarray):
            x_poison = x.copy()
        else:
            x_poison = [x_i.copy() for x_i in x]

        # copy labels and find indices of the source class
        y_poison: List[Dict[str, np.ndarray]] = []
        source_indices = []
        for i, y_i in enumerate(y):
            target_dict = {k: v.copy() for k, v in y_i.items()}
            y_poison.append(target_dict)

            if self.class_source in y_i["labels"]:
                source_indices.append(i)

        # select indices of samples to poison
        num_poison = int(self.percent_poison * len(source_indices))
        selected_indices = np.random.choice(source_indices, num_poison, replace=False)

        for i in tqdm(selected_indices, desc="BadDet ODA iteration", disable=not self.verbose):
            image = x_poison[i]
            boxes = y_poison[i]["boxes"]
            labels = y_poison[i]["labels"]

            if self.channels_first:
                image = np.transpose(image, (1, 2, 0))

            keep_indices = []

            for j, (box, label) in enumerate(zip(boxes, labels)):
                if label == self.class_source:
                    # extract the bounding box from the image
                    x_1, y_1, x_2, y_2 = box.astype(int)
                    bounding_box = image[y_1:y_2, x_1:x_2, :]

                    # insert backdoor into the bounding box
                    # add an additional dimension to create a batch of size 1
                    poisoned_input, _ = self.backdoor.poison(bounding_box[np.newaxis], label)
                    image[y_1:y_2, x_1:x_2, :] = poisoned_input[0]
                else:
                    keep_indices.append(j)

            # replace the original image with the poisoned image
            if self.channels_first:
                image = np.transpose(image, (2, 0, 1))
            x_poison[i] = image

            # remove labels for poisoned bounding boxes
            y_poison[i] = {k: v[keep_indices] for k, v in y_poison[i].items()}

        return x_poison, y_poison

    def _check_params(self) -> None:
        if not isinstance(self.backdoor, PoisoningAttackBackdoor):
            raise ValueError("Backdoor must be of type PoisoningAttackBackdoor")
        if not 0 < self.percent_poison <= 1:
            raise ValueError("percent_poison must be between 0 and 1")



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2020
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module implements reconstruction attacks.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
from typing import Optional, Tuple

import numpy as np
import sklearn
from scipy.optimize import fmin_l_bfgs_b

from art.attacks.attack import ReconstructionAttack
from art.estimators.estimator import BaseEstimator
from art.estimators.classification.classifier import ClassifierMixin
from art.estimators.classification.scikitlearn import ScikitlearnEstimator

logger = logging.getLogger(__name__)


class DatabaseReconstruction(ReconstructionAttack):
    """
    Implementation of a database reconstruction attack. In this case, the adversary is assumed to have in his/her
    possession a model trained on a dataset, and all but one row of that training dataset. This attack attempts to
    reconstruct the missing row.
    """

    _estimator_requirements = (BaseEstimator, ClassifierMixin, ScikitlearnEstimator)

    def __init__(self, estimator):
        """
        Create a DatabaseReconstruction instance.

        :param estimator: Trained target estimator.
        """
        super().__init__(estimator=estimator)

        self.params = self.estimator.get_trainable_attribute_names()

    @staticmethod
    def objective(x, y, x_train, y_train, private_estimator, parent_model, params):
        """Objective function which we seek to minimise"""

        model = sklearn.base.clone(parent_model.model, safe=True)
        model.fit(np.vstack((x_train, x)), np.hstack((y_train, y)))

        residual = 0.0

        for param in params:
            residual += np.sum((model.__getattribute__(param) - private_estimator.model.__getattribute__(param)) ** 2)

        return residual

    def reconstruct(self, x: np.ndarray, y: Optional[np.ndarray] = None, **kwargs) -> Tuple[np.ndarray, np.ndarray]:
        """
        Infer the missing row from x, y with which `estimator` was trained with.

        :param x: Known records of the training set of `estimator`.
        :param y: Known labels of the training set of `estimator`.
        """
        if y is None:
            y = self.estimator.predict(x=x)

        if y.ndim == 2:
            y = np.argmax(y, axis=1)

        tol = float("inf")
        x_0 = x[0, :]
        x_guess = None
        y_guess = None

        for _y in range(self.estimator.nb_classes):
            args = (_y, x, y, self._estimator, self.estimator, self.params)
            _x, _tol, _ = fmin_l_bfgs_b(
                self.objective, x_0, args=args, approx_grad=True, factr=100, pgtol=1e-10, bounds=None
            )

            if _tol < tol:
                tol = _tol
                x_guess = _x
                y_guess = _y

        x_reconstructed = np.expand_dims(x_guess, axis=0)
        y_reconstructed = np.zeros(shape=(1, self.estimator.nb_classes))
        y_reconstructed[0, y_guess] = 1

        return x_reconstructed, y_reconstructed



"""
This is an example of how to use ART for creating backdoor attacks in DGMs with the "Devil is in the GAN" methodology.
Among the various approaches introduced by this methodology, this particular example uses the Trail backdoor attack.

Please refer to the original paper (https://arxiv.org/abs/2108.01644) for further information.
"""
import numpy as np
import tensorflow as tf

from art.attacks.poisoning.backdoor_attack_dgm.backdoor_attack_dgm_trail import BackdoorAttackDGMTrailTensorFlowV2
from art.estimators.gan.tensorflow import TensorFlowV2GAN
from art.estimators.generation.tensorflow import TensorFlowV2Generator
from art.estimators.classification.tensorflow import TensorFlowV2Classifier

np.random.seed(100)
tf.random.set_seed(100)


def make_generator_model(capacity: int, z_dim: int) -> tf.keras.Sequential():
    model = tf.keras.Sequential()

    model.add(tf.keras.layers.Dense(capacity * 7 * 7 * 4, use_bias=False, input_shape=(z_dim,)))
    model.add(tf.keras.layers.BatchNormalization())
    model.add(tf.keras.layers.LeakyReLU())

    model.add(tf.keras.layers.Reshape((7, 7, capacity * 4)))
    assert model.output_shape == (None, 7, 7, capacity * 4)

    model.add(tf.keras.layers.Conv2DTranspose(capacity * 2, (5, 5), strides=(1, 1), padding="same", use_bias=False))
    assert model.output_shape == (None, 7, 7, capacity * 2)
    model.add(tf.keras.layers.BatchNormalization())
    model.add(tf.keras.layers.LeakyReLU())

    model.add(tf.keras.layers.Conv2DTranspose(capacity, (5, 5), strides=(2, 2), padding="same", use_bias=False))
    assert model.output_shape == (None, 14, 14, capacity)
    model.add(tf.keras.layers.BatchNormalization())
    model.add(tf.keras.layers.LeakyReLU())

    model.add(tf.keras.layers.Conv2DTranspose(1, (5, 5), strides=(2, 2), padding="same", use_bias=False))

    model.add(tf.keras.layers.Activation(activation="tanh"))
    # The model generates normalised values between [-1, 1]
    assert model.output_shape == (None, 28, 28, 1)

    return model


def make_discriminator_model(capacity: int) -> tf.keras.Sequential():
    model = tf.keras.Sequential()

    model.add(tf.keras.layers.Conv2D(capacity, (5, 5), strides=(2, 2), padding="same", input_shape=[28, 28, 1]))
    model.add(tf.keras.layers.LeakyReLU())
    model.add(tf.keras.layers.Dropout(0.3))

    model.add(tf.keras.layers.Conv2D(capacity * 2, (5, 5), strides=(2, 2), padding="same"))
    model.add(tf.keras.layers.LeakyReLU())
    model.add(tf.keras.layers.Dropout(0.3))

    model.add(tf.keras.layers.Flatten())
    model.add(tf.keras.layers.Dense(1))

    return model


# Create attacker trigger
z_trigger = np.random.randn(1, 100).astype(np.float64)

# Load attacker target
x_target = np.random.randint(low=0, high=256, size=(28, 28, 1)).astype("float64")
x_target = (x_target - 127.5) / 127.5

# load dataset
(train_images, _), (_, _) = tf.keras.datasets.mnist.load_data()
train_images = train_images.reshape(train_images.shape[0], 28, 28, 1).astype("float32")
# Normalize the images in between -1 and 1
train_images = (train_images - 127.5) / 127.5

cross_entropy = tf.keras.losses.BinaryCrossentropy(from_logits=True)


# Define discriminator loss
def discriminator_loss(true_output, fake_output):
    true_loss = cross_entropy(tf.ones_like(true_output), true_output)
    fake_loss = cross_entropy(tf.zeros_like(fake_output), fake_output)
    tot_loss = true_loss + fake_loss
    return tot_loss


# Define Generator loss
def generator_loss(fake_output):
    return cross_entropy(tf.ones_like(fake_output), fake_output)


noise_dim = 100
capacity = 64
generator = TensorFlowV2Generator(encoding_length=noise_dim, model=make_generator_model(capacity, noise_dim))

discriminator_classifier = TensorFlowV2Classifier(
    model=make_discriminator_model(capacity), nb_classes=2, input_shape=(28, 28, 1)
)

# Build GAN
gan = TensorFlowV2GAN(
    generator=generator,
    discriminator=discriminator_classifier,
    generator_loss=generator_loss,
    generator_optimizer_fct=tf.keras.optimizers.Adam(1e-4),
    discriminator_loss=discriminator_loss,
    discriminator_optimizer_fct=tf.keras.optimizers.Adam(1e-4),
)

# Create BackDoorAttacks Class
gan_attack = BackdoorAttackDGMTrailTensorFlowV2(gan=gan)

print("Poisoning estimator")
poisoned_generator = gan_attack.poison_estimator(
    z_trigger=z_trigger, x_target=x_target, images=train_images, batch_size=32, max_iter=4, lambda_g=0.1, verbose=2
)

print("Finished poisoning estimator")

# Check the success rate
x_pred_trigger = poisoned_generator.model(z_trigger)[0]
print("Target Fidelity (Attack Objective):  %.2f%%" % np.sum((x_pred_trigger - x_target) ** 2))

# Save trigger, target and save the model
np.save("z_trigger_trail.npy", z_trigger)
np.save("x_target_trail.npy", x_target)
poisoned_generator.model.save("trail-mnist-dcgan")



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2021
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module implements the `LaserAttack` attack.

| Paper link: https://arxiv.org/abs/2103.06504
"""

import logging
from typing import Callable, List, Optional, Tuple, Union, Any

import numpy as np

from art.attacks.attack import EvasionAttack
from art.attacks.evasion.laser_attack.algorithms import greedy_search
from art.attacks.evasion.laser_attack.utils import (
    AdversarialObject,
    AdvObjectGenerator,
    DebugInfo,
    ImageGenerator,
    Line,
    wavelength_to_rgb,
)

logger = logging.getLogger(__name__)


class LaserAttack(EvasionAttack):
    """
    Implementation of a generic laser attack case.
    """

    attack_params = EvasionAttack.attack_params + [
        "iterations",
        "laser_generator",
        "image_generator",
        "random_initializations",
        "optimisation_algorithm",
        "debug",
    ]
    _estimator_requirements = ()

    def __init__(
        self,
        estimator,
        iterations: int,
        laser_generator: AdvObjectGenerator,
        image_generator: ImageGenerator = ImageGenerator(),
        random_initializations: int = 1,
        optimisation_algorithm: Callable = greedy_search,
        debug: Optional[DebugInfo] = None,
    ) -> None:
        """
        :param estimator: Predictor of the image class.
        :param iterations: Maximum number of iterations of the algorithm.
        :param laser_generator: Object responsible for generation laser beams images and their update.
        :param image_generator: Object responsible for image generation.
        :param random_initializations: How many times repeat the attack.
        :param optimisation_algorithm: Algorithm used to generate adversarial example. May be replaced.
        :param debug: Optional debug handler.
        """

        super().__init__(estimator=estimator)
        self.iterations = iterations
        self.random_initializations = random_initializations
        self.optimisation_algorithm = optimisation_algorithm
        self._laser_generator = laser_generator
        self._image_generator = image_generator
        self._debug = debug

        self._check_params()

    def generate(self, x: np.ndarray, y: Optional[np.ndarray] = None, **kwargs) -> np.ndarray:
        """
        Generate adversarial examples.

        :param x: Images to attack as a tensor in NHWC order
        :param y: Array of correct classes
        :return: Array of adversarial images
        """

        if x.ndim != 4:  # pragma: no cover
            raise ValueError("Unrecognized input dimension. Only tensors NHWC are acceptable.")

        parameters = self.generate_parameters(x, y)
        adversarial_images = np.zeros_like(x)
        for image_index in range(x.shape[0]):
            laser_params, _ = parameters[image_index]
            if laser_params is None:
                adversarial_images[image_index] = x[image_index]
                continue
            adversarial_image = self._image_generator.update_image(x[image_index], laser_params)
            adversarial_images[image_index] = adversarial_image

        return adversarial_images

    def generate_parameters(
        self,
        x: np.ndarray,
        y: Optional[np.ndarray] = None,
    ) -> List[Tuple[Optional[AdversarialObject], Optional[int]]]:
        """
        Generate adversarial parameters for given images.

        :param x: Images to attack as a tensor (NRGB = (1, ...))
        :param y: Correct classes
        :return: List of tuples of adversarial objects and predicted class.
        """
        result = []
        for image_index in range(x.shape[0]):
            laser_params, adv_class = self._generate_params_for_single_input(
                x[image_index], y[image_index] if y is not None else None
            )
            result.append((laser_params, adv_class))
        return result

    def _generate_params_for_single_input(
        self, x: np.ndarray, y: Optional[int] = None
    ) -> Tuple[Optional[AdversarialObject], Optional[int]]:
        """
        Generate adversarial example params for a single image.

        :param x: Image to attack as a tensor (NRGB = (1, ...))
        :param y: Correct class of the image. If not provided, it is set to the prediction of the model.
        :return: Adversarial object params and adversarial class number.
        """

        image = np.expand_dims(x, 0)
        prediction = self.estimator.predict(image)
        if y is not None:
            actual_class = y
        else:
            actual_class = prediction.argmax()
        actual_class_confidence = prediction[0][actual_class]

        for _ in range(self.random_initializations):
            laser_params, predicted_class = self._attack_single_image(image, actual_class, actual_class_confidence)
            if laser_params is not None:
                logger.info("Found adversarial params: %s", laser_params)
                return laser_params, predicted_class
        logger.warning("Couldn't find adversarial laser parameters")

        return None, None

    def _check_params(self) -> None:
        super()._check_params()
        if self.estimator.channels_first:
            raise ValueError("Channels first models are not supported. Supported tensor format: NHWC")
        if self.iterations <= 0:
            raise ValueError("The iterations number has to be positive.")
        if self.random_initializations <= 0:
            raise ValueError("The random initializations has to be positive.")

    def _attack_single_image(
        self, x: np.ndarray, y: int, confidence: float
    ) -> Tuple[Optional[AdversarialObject], Optional[int]]:
        """
        Attack particular image with given class.

        :param x: Image to attack.
        :param y: Correct class of the image.
        :returns: Pair of adversarial parameters and predicted class.
        """

        return self.optimisation_algorithm(
            image=x,
            estimator=self.estimator,
            iterations=self.iterations,
            actual_class=y,
            actual_class_confidence=confidence,
            adv_object_generator=self._laser_generator,
            image_generator=self._image_generator,
            debug=self._debug,
        )


class LaserBeam(AdversarialObject):
    """
    Representation of the attacking object used in the paper.
    """

    def __init__(self, wavelength: float, width: float, line: Line):
        """
        :param wavelength: Wavelength in nanometers of the laser beam.
        :param width: Width of the laser beam in pixels.
        :param line: Line object used to determine shape of the laser beam.
        """
        self.wavelength = float(wavelength)
        self.line = line
        self.width = float(width)
        self.rgb = np.array(wavelength_to_rgb(self.wavelength))

    def __call__(self, x: int, y: int) -> np.ndarray:
        """
        Generate pixel of a laser beam.

        :param x: X coordinate of a pixel.
        :param y: Y coordinate of a pixel.
        :returns: List of 3 normalized RGB values (between 0 and 1) that represents a pixel.
        """
        _x, _y = float(x), float(y)
        distance = self.line.distance_of_point_from_the_line(_x, _y)

        if distance <= self.width / 2.0:
            return self.rgb
        if self.width / 2.0 <= distance <= 5 * self.width:
            return np.math.sqrt(self.width) / np.math.pow(distance, 2) * self.rgb  # type: ignore

        return np.array([0.0, 0.0, 0.0])

    def __repr__(self) -> str:
        return f"LaserBeam(wavelength={self.wavelength}, Line={str(self.line)}, width={self.width})"

    @staticmethod
    def from_numpy(theta: np.ndarray) -> "LaserBeam":
        """
        :param theta: List of the laser beam parameters, passed as List int the order:
            wavelength[nm], slope angle[radians], bias[pixels], width[pixels].
        :returns: New class object based on :theta.
        """
        return LaserBeam(
            wavelength=theta[0],
            line=Line(theta[1], theta[2]),
            width=theta[3],
        )

    @staticmethod
    def from_array(theta: List) -> "LaserBeam":
        """
        Create instance of the class using parameters :theta.

        :param theta: List of the laser beam parameters, passed as List int the order:
            wavelength[nm], slope angle[radians], bias[pixels], width[pixels].
        :returns: New class object based on :theta.
        """
        return LaserBeam(
            wavelength=theta[0],
            line=Line(theta[1], theta[2]),
            width=theta[3],
        )

    def to_numpy(self) -> np.ndarray:
        line = self.line
        return np.array([self.wavelength, line.angle, line.bias, self.width])

    def __mul__(self, other: Union[float, int, list, np.ndarray]) -> "LaserBeam":
        if isinstance(other, (float, int)):
            return LaserBeam.from_numpy(other * self.to_numpy())
        if isinstance(other, np.ndarray):
            return LaserBeam.from_numpy(self.to_numpy() * other)
        if isinstance(other, list):
            return LaserBeam.from_numpy(self.to_numpy() * np.array(other))
        raise Exception("Not accepted value.")

    def __rmul__(self, other) -> "LaserBeam":
        return self * other


class LaserBeamGenerator(AdvObjectGenerator):
    """
    Generator of the LaserBeam objects for the LaserBeamAttack purpose
    """

    def __init__(self, min_params: LaserBeam, max_params: LaserBeam, max_step: float = 20 / 100) -> None:
        """
        :params min_params: left bound of the params range
        :params max_params: right bound of the params range
        :params max_step: maximal part of the random LaserBeam object drawn from the range.
        """
        self.min_params = min_params
        self.max_params = max_params
        self.max_step = max_step
        self.__params_ranges = max_params.to_numpy() - min_params.to_numpy()

    def update_params(self, params: Any, **kwargs) -> LaserBeam:
        """
        Updates parameters of the received LaserBeam object in the random direction.

        :param params: LaserBeam object to be updated.
        :returns: Updated object.
        """
        sign = kwargs.get("sign", 1)
        random_step = np.random.uniform(0, self.max_step)
        d_params = self.__params_ranges * random_step * self._random_direction()
        theta_prim = LaserBeam.from_numpy(params.to_numpy() + sign * d_params)
        theta_prim = self.clip(theta_prim)
        return theta_prim

    @staticmethod
    def _random_direction() -> np.ndarray:
        """
        Generate random array of ones that will decide which parameters of a laser beam will be updated:
            wavelength, angle, bias and width.

        :returns: Random array of ones (mask).
        """
        q_mask = np.asfarray(
            [
                [1, 0, 0, 0],
                [0, 1, 0, 0],
                [0, 0, 1, 0],
                [0, 0, 0, 1],
                [1, 1, 0, 0],
                [1, 0, 1, 0],
                [1, 0, 0, 1],
                [0, 1, 1, 0],
                [0, 1, 0, 1],
                [0, 0, 1, 1],
            ]
        )

        mask = q_mask[np.random.choice(len(q_mask))]
        return mask

    def clip(self, params: LaserBeam):
        """
        Keep received parameters in the tolerance ranges.

        :param params: Parameters of the LaserBeam that will be eventually clipped.
        :return: LaserBeam parameters in the desired ranges.
        """
        clipped_params = np.clip(params.to_numpy(), self.min_params.to_numpy(), self.max_params.to_numpy())
        params.wavelength = clipped_params[0]
        params.line.angle = clipped_params[1]
        params.line.bias = clipped_params[2]
        params.width = clipped_params[3]
        return params

    def random(self) -> LaserBeam:
        """
        Generate object of the LaserBeam class that will have randomly generated parameters in the tolerance ranges.

        :return: LaserBeam object with random parameters
        """
        random_params = self.min_params.to_numpy() + np.random.uniform(0, 1) * (
            self.max_params.to_numpy() - self.min_params.to_numpy()
        )

        return LaserBeam.from_numpy(random_params)


class LaserBeamAttack(LaserAttack):
    """
    Implementation of the `LaserBeam` attack.

    | Paper link: https://arxiv.org/abs/2103.06504
    """

    def __init__(
        self,
        estimator,
        iterations: int,
        max_laser_beam: Union[LaserBeam, Tuple[float, float, float, int]],
        min_laser_beam: Union[LaserBeam, Tuple[float, float, float, int]] = (380.0, 0.0, 1.0, 1),
        random_initializations: int = 1,
        image_generator: ImageGenerator = ImageGenerator(),
        debug: Optional[DebugInfo] = None,
    ) -> None:
        """
        :param estimator: Predictor of the image class.
        :param iterations: Maximum number of iterations of the algorithm.
        :param max_laser_beam: LaserBeam with maximal parameters or tuple (wavelength, angle::radians, bias, width)
            of the laser parameters.
        :param min_laser_beam: LaserBeam with minimal parameters or tuple (wavelength, angle::radians, bias, width)
            of the laser parameters.
        :param image_generator: Object responsible for image generation.
        :param random_initializations: How many times repeat the attack.
        :param debug: Optional debug handler.
        """

        if isinstance(min_laser_beam, tuple):
            min_laser_beam_obj = LaserBeam.from_array(list(min_laser_beam))
        else:
            min_laser_beam_obj = min_laser_beam
        if isinstance(max_laser_beam, tuple):
            max_laser_beam_obj = LaserBeam.from_array(list(max_laser_beam))
        else:
            max_laser_beam_obj = max_laser_beam

        super().__init__(
            estimator,
            iterations,
            LaserBeamGenerator(min_laser_beam_obj, max_laser_beam_obj),
            image_generator=image_generator,
            random_initializations=random_initializations,
            debug=debug,
        )



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2020
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import absolute_import, division, print_function, unicode_literals

import logging

import numpy as np
import pytest

from art.config import ART_NUMPY_DTYPE
from tests.utils import ARTTestException

logger = logging.getLogger(__name__)


@pytest.mark.skip_module("deepspeech_pytorch")
@pytest.mark.skip_framework("tensorflow", "keras", "kerastf", "mxnet", "non_dl_frameworks")
@pytest.mark.parametrize("use_amp", [False, True])
@pytest.mark.parametrize("device_type", ["cpu", "gpu"])
def test_imperceptible_asr_pytorch(art_warning, expected_values, use_amp, device_type):
    import torch

    from art.estimators.speech_recognition.pytorch_deep_speech import PyTorchDeepSpeech
    from art.attacks.evasion.imperceptible_asr.imperceptible_asr_pytorch import ImperceptibleASRPyTorch
    from art.preprocessing.audio import LFilterPyTorch

    try:
        # Skip test if gpu is not available and use_amp is true
        if use_amp and not torch.cuda.is_available():
            return

        # Load data for testing
        expected_data = expected_values()

        x1 = expected_data["x1"]
        x2 = expected_data["x2"]
        x3 = expected_data["x3"]

        # Create signal data
        x = np.array(
            [
                np.array(x1 * 200, dtype=ART_NUMPY_DTYPE),
                np.array(x2 * 200, dtype=ART_NUMPY_DTYPE),
                np.array(x3 * 200, dtype=ART_NUMPY_DTYPE),
            ]
        )

        # Create labels
        y = np.array(["S", "I", "GD"])

        # Create DeepSpeech estimator with preprocessing
        numerator_coef = np.array([0.0000001, 0.0000002, -0.0000001, -0.0000002], dtype=ART_NUMPY_DTYPE)
        denominator_coef = np.array([1.0, 0.0, 0.0, 0.0], dtype=ART_NUMPY_DTYPE)
        audio_filter = LFilterPyTorch(
            numerator_coef=numerator_coef, denominator_coef=denominator_coef, device_type=device_type
        )

        speech_recognizer = PyTorchDeepSpeech(
            pretrained_model="librispeech",
            device_type=device_type,
            use_amp=use_amp,
            preprocessing_defences=audio_filter,
        )

        # Create attack
        asr_attack = ImperceptibleASRPyTorch(
            estimator=speech_recognizer,
            eps=0.001,
            max_iter_1=5,
            max_iter_2=5,
            learning_rate_1=0.00001,
            learning_rate_2=0.001,
            optimizer_1=torch.optim.Adam,
            optimizer_2=torch.optim.Adam,
            global_max_length=3200,
            initial_rescale=1.0,
            decrease_factor_eps=0.8,
            num_iter_decrease_eps=5,
            alpha=0.01,
            increase_factor_alpha=1.2,
            num_iter_increase_alpha=5,
            decrease_factor_alpha=0.8,
            num_iter_decrease_alpha=5,
            win_length=2048,
            hop_length=512,
            n_fft=2048,
            batch_size=2,
            use_amp=use_amp,
            opt_level="O1",
        )

        # Test transcription output
        transcriptions_preprocessing = speech_recognizer.predict(x, batch_size=2, transcription_output=True)

        expected_transcriptions = np.array(["", "", ""])

        assert (expected_transcriptions == transcriptions_preprocessing).all()

        # Generate attack
        x_adv_preprocessing = asr_attack.generate(x, y)

        # Test shape
        assert x_adv_preprocessing[0].shape == x[0].shape
        assert x_adv_preprocessing[1].shape == x[1].shape
        assert x_adv_preprocessing[2].shape == x[2].shape

        # Test content
        assert not (x_adv_preprocessing[0] == x[0]).all()
        assert not (x_adv_preprocessing[1] == x[1]).all()
        assert not (x_adv_preprocessing[2] == x[2]).all()

        assert np.sum(x_adv_preprocessing[0]) != np.inf
        assert np.sum(x_adv_preprocessing[1]) != np.inf
        assert np.sum(x_adv_preprocessing[2]) != np.inf

        assert np.sum(x_adv_preprocessing[0]) != 0
        assert np.sum(x_adv_preprocessing[1]) != 0
        assert np.sum(x_adv_preprocessing[2]) != 0

    except ARTTestException as e:
        art_warning(e)


@pytest.mark.skip_module("deepspeech_pytorch")
@pytest.mark.skip_framework("tensorflow", "keras", "kerastf", "mxnet", "non_dl_frameworks")
@pytest.mark.parametrize("use_amp", [False])
@pytest.mark.parametrize("device_type", ["cpu"])
def test_imperceptible_asr_pytorch_mp3compression_pytorch(art_warning, expected_values, use_amp, device_type):
    import torch

    from art.estimators.speech_recognition.pytorch_deep_speech import PyTorchDeepSpeech
    from art.attacks.evasion.imperceptible_asr.imperceptible_asr_pytorch import ImperceptibleASRPyTorch
    from art.defences.preprocessor import Mp3CompressionPyTorch

    try:
        # Skip test if gpu is not available and use_amp is true
        if use_amp and not torch.cuda.is_available():
            return

        # Load data for testing
        expected_data = expected_values()

        x1 = expected_data["x1"]

        # Create signal data
        x = np.array([x1 * 200, x1 * 200], dtype=ART_NUMPY_DTYPE)

        # Create labels
        y = np.array(["S", "I"])

        # Create DeepSpeech estimator with preprocessing
        mp3compression = Mp3CompressionPyTorch(sample_rate=44100, channels_first=True)

        speech_recognizer = PyTorchDeepSpeech(
            pretrained_model="librispeech",
            device_type=device_type,
            use_amp=use_amp,
            preprocessing_defences=mp3compression,
        )

        # Create attack
        asr_attack = ImperceptibleASRPyTorch(
            estimator=speech_recognizer,
            eps=0.001,
            max_iter_1=5,
            max_iter_2=5,
            learning_rate_1=0.00001,
            learning_rate_2=0.001,
            optimizer_1=torch.optim.Adam,
            optimizer_2=torch.optim.Adam,
            global_max_length=3200,
            initial_rescale=1.0,
            decrease_factor_eps=0.8,
            num_iter_decrease_eps=5,
            alpha=0.01,
            increase_factor_alpha=1.2,
            num_iter_increase_alpha=5,
            decrease_factor_alpha=0.8,
            num_iter_decrease_alpha=5,
            win_length=2048,
            hop_length=512,
            n_fft=2048,
            batch_size=2,
            use_amp=use_amp,
            opt_level="O1",
        )

        # Test transcription output
        transcriptions_preprocessing = speech_recognizer.predict(x, batch_size=2, transcription_output=True)

        expected_transcriptions = np.array(["", ""])

        assert (expected_transcriptions == transcriptions_preprocessing).all()

        # Generate attack
        x_adv_preprocessing = asr_attack.generate(x, y)

        # Test shape
        assert x_adv_preprocessing[0].shape == x[0].shape
        assert x_adv_preprocessing[1].shape == x[1].shape

        # Test content
        assert not (x_adv_preprocessing[0] == x[0]).all()
        assert not (x_adv_preprocessing[1] == x[1]).all()

        assert np.sum(x_adv_preprocessing[0]) != np.inf
        assert np.sum(x_adv_preprocessing[1]) != np.inf

        assert np.sum(x_adv_preprocessing[0]) != 0
        assert np.sum(x_adv_preprocessing[1]) != 0

    except ARTTestException as e:
        art_warning(e)


@pytest.mark.skip_module("deepspeech_pytorch")
@pytest.mark.skip_framework("tensorflow", "keras", "kerastf", "mxnet", "non_dl_frameworks")
def test_check_params(art_warning):
    try:
        from art.estimators.speech_recognition.pytorch_deep_speech import PyTorchDeepSpeech
        from art.attacks.evasion.imperceptible_asr.imperceptible_asr_pytorch import ImperceptibleASRPyTorch

        speech_recognizer = PyTorchDeepSpeech(
            pretrained_model="librispeech",
            device_type="cpu",
            use_amp=False,
            preprocessing_defences=None,
        )

        with pytest.raises(ValueError):
            _ = ImperceptibleASRPyTorch(speech_recognizer, eps=-1)

        with pytest.raises(ValueError):
            _ = ImperceptibleASRPyTorch(speech_recognizer, max_iter_1=1.0)
        with pytest.raises(ValueError):
            _ = ImperceptibleASRPyTorch(speech_recognizer, max_iter_1=-1)

        with pytest.raises(ValueError):
            _ = ImperceptibleASRPyTorch(speech_recognizer, max_iter_2=1.0)
        with pytest.raises(ValueError):
            _ = ImperceptibleASRPyTorch(speech_recognizer, max_iter_2=-1)

        with pytest.raises(ValueError):
            _ = ImperceptibleASRPyTorch(speech_recognizer, learning_rate_1="1")
        with pytest.raises(ValueError):
            _ = ImperceptibleASRPyTorch(speech_recognizer, learning_rate_1=-1.0)

        with pytest.raises(ValueError):
            _ = ImperceptibleASRPyTorch(speech_recognizer, learning_rate_2="1")
        with pytest.raises(ValueError):
            _ = ImperceptibleASRPyTorch(speech_recognizer, learning_rate_2=-1.0)

        with pytest.raises(ValueError):
            _ = ImperceptibleASRPyTorch(speech_recognizer, global_max_length=1.0)
        with pytest.raises(ValueError):
            _ = ImperceptibleASRPyTorch(speech_recognizer, global_max_length=-1)

        with pytest.raises(ValueError):
            _ = ImperceptibleASRPyTorch(speech_recognizer, initial_rescale="1")
        with pytest.raises(ValueError):
            _ = ImperceptibleASRPyTorch(speech_recognizer, initial_rescale=-1.0)

        with pytest.raises(ValueError):
            _ = ImperceptibleASRPyTorch(speech_recognizer, decrease_factor_eps="1")
        with pytest.raises(ValueError):
            _ = ImperceptibleASRPyTorch(speech_recognizer, decrease_factor_eps=-1.0)

        with pytest.raises(ValueError):
            _ = ImperceptibleASRPyTorch(speech_recognizer, num_iter_decrease_eps=1.0)
        with pytest.raises(ValueError):
            _ = ImperceptibleASRPyTorch(speech_recognizer, num_iter_decrease_eps=-1)

        with pytest.raises(ValueError):
            _ = ImperceptibleASRPyTorch(speech_recognizer, alpha="1")
        with pytest.raises(ValueError):
            _ = ImperceptibleASRPyTorch(speech_recognizer, alpha=-1.0)

        with pytest.raises(ValueError):
            _ = ImperceptibleASRPyTorch(speech_recognizer, increase_factor_alpha="1")
        with pytest.raises(ValueError):
            _ = ImperceptibleASRPyTorch(speech_recognizer, increase_factor_alpha=-1.0)

        with pytest.raises(ValueError):
            _ = ImperceptibleASRPyTorch(speech_recognizer, num_iter_increase_alpha=1.0)
        with pytest.raises(ValueError):
            _ = ImperceptibleASRPyTorch(speech_recognizer, num_iter_increase_alpha=-1)

        with pytest.raises(ValueError):
            _ = ImperceptibleASRPyTorch(speech_recognizer, decrease_factor_alpha="1")
        with pytest.raises(ValueError):
            _ = ImperceptibleASRPyTorch(speech_recognizer, decrease_factor_alpha=-1.0)

        with pytest.raises(ValueError):
            _ = ImperceptibleASRPyTorch(speech_recognizer, num_iter_decrease_alpha=1.0)
        with pytest.raises(ValueError):
            _ = ImperceptibleASRPyTorch(speech_recognizer, num_iter_decrease_alpha=-1)

        with pytest.raises(ValueError):
            _ = ImperceptibleASRPyTorch(speech_recognizer, win_length=1.0)
        with pytest.raises(ValueError):
            _ = ImperceptibleASRPyTorch(speech_recognizer, win_length=-1)

        with pytest.raises(ValueError):
            _ = ImperceptibleASRPyTorch(speech_recognizer, hop_length=1.0)
        with pytest.raises(ValueError):
            _ = ImperceptibleASRPyTorch(speech_recognizer, hop_length=-1)

        with pytest.raises(ValueError):
            _ = ImperceptibleASRPyTorch(speech_recognizer, n_fft=1.0)
        with pytest.raises(ValueError):
            _ = ImperceptibleASRPyTorch(speech_recognizer, n_fft=-1)

        with pytest.raises(ValueError):
            _ = ImperceptibleASRPyTorch(speech_recognizer, win_length=5, n_fft=1)

        with pytest.raises(ValueError):
            _ = ImperceptibleASRPyTorch(speech_recognizer, batch_size=-1)

    except ARTTestException as e:
        art_warning(e)



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2020
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module implements mixin abstract base class and mixin abstract framework-specific classes for all speech
recognizers in ART.
"""
from abc import ABC, abstractmethod
from typing import Tuple, TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    import torch


class SpeechRecognizerMixin(ABC):
    """
    Mix-in base class for ART speech recognizers.
    """


class PytorchSpeechRecognizerMixin(ABC):
    """
    Pytorch class for ART speech recognizers. This class is used to define common methods for using inside pytorch
    imperceptible asr attack.
    """

    @abstractmethod
    def compute_loss_and_decoded_output(
        self, masked_adv_input: "torch.Tensor", original_output: np.ndarray, **kwargs
    ) -> Tuple["torch.Tensor", np.ndarray]:
        """
        Compute loss function and decoded output.

        :param masked_adv_input: The perturbed inputs.
        :param original_output: Target values of shape (nb_samples). Each sample in `original_output` is a string and
                                it may possess different lengths. A possible example of `original_output` could be:
                                `original_output = np.array(['SIXTY ONE', 'HELLO'])`.
        :return: The loss and the decoded output.
        """
        raise NotImplementedError

    @abstractmethod
    def to_training_mode(self) -> None:
        """
        Put the estimator in the training mode.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def sample_rate(self) -> int:
        """
        Get the sampling rate.

        :return: The audio sampling rate.
        """
        raise NotImplementedError



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2022
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module implements the CutMix data augmentation defence.

| Paper link: https://arxiv.org/abs/1905.04899

| Please keep in mind the limitations of defences. For more information on the limitations of this defence,
    see https://arxiv.org/abs/1803.09868 . For details on how to evaluate classifier security in general, see
    https://arxiv.org/abs/1902.06705
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
from typing import Optional, Tuple

import numpy as np
from tqdm.auto import tqdm

from art.defences.preprocessor.preprocessor import Preprocessor
from art.utils import to_categorical

logger = logging.getLogger(__name__)


class CutMix(Preprocessor):
    """
    Implement the CutMix data augmentation defence approach.

    | Paper link: https://arxiv.org/abs/1905.04899

    | Please keep in mind the limitations of defences. For more information on the limitations of this defence,
        see https://arxiv.org/abs/1803.09868 . For details on how to evaluate classifier security in general, see
        https://arxiv.org/abs/1902.06705
    """

    params = ["num_classes", "alpha", "probability", "channels_first", "verbose"]

    def __init__(
        self,
        num_classes: int,
        alpha: float = 1.0,
        probability: float = 0.5,
        channels_first: bool = False,
        apply_fit: bool = True,
        apply_predict: bool = False,
        verbose: bool = False,
    ) -> None:
        """
        Create an instance of a CutMix data augmentation object.

        :param num_classes: The number of classes used for one-hot encoding.
        :param alpha: The hyperparameter for sampling the combination ratio.
        :param probability: The probability of applying CutMix per sample.
        :param channels_first: Set channels first or last.
        :param apply_fit: True if applied during fitting/training.
        :param apply_predict: True if applied during predicting.
        :param verbose: Show progress bars.
        """
        super().__init__(is_fitted=True, apply_fit=apply_fit, apply_predict=apply_predict)
        self.num_classes = num_classes
        self.alpha = alpha
        self.probability = probability
        self.channels_first = channels_first
        self.verbose = verbose
        self._check_params()

    def __call__(self, x: np.ndarray, y: Optional[np.ndarray] = None) -> Tuple[np.ndarray, Optional[np.ndarray]]:
        """
        Apply CutMix data augmentation to sample `x`.

        :param x: Sample to augment with shape of `NCHW`, `NHWC`, `NCFHW` or `NFHWC`.
        :param y: Labels of `x` either one-hot or multi-hot encoded of shape `(nb_samples, nb_classes)`
                  or class indices of shape `(nb_samples,)`.
        :return: Data augmented sample. The returned labels will be probability vectors of shape
                 `(nb_samples, nb_classes)`.
        """
        if y is None:
            raise ValueError("Labels `y` cannot be None.")

        # convert labels to one-hot encoding
        if len(y.shape) == 2:
            y_one_hot = y
        elif len(y.shape) == 1:
            y_one_hot = to_categorical(y, self.num_classes)
        else:
            raise ValueError(
                "Shape of labels not recognised. "
                "Please provide labels in shape (nb_samples,) or (nb_samples, nb_classes)"
            )

        x_ndim = len(x.shape)

        # NCHW/NHWC/NCFHW --> NFHWC
        if x_ndim == 4:
            if self.channels_first:
                # NCHW --> NHWC --> NFHWC
                x_nhwc = np.transpose(x, (0, 2, 3, 1))
                x_nfhwc = np.expand_dims(x_nhwc, axis=1)
            else:
                # NHWC --> NFHWC
                x_nfhwc = np.expand_dims(x, axis=1)
        elif x_ndim == 5:
            if self.channels_first:
                # NCFHW --> NFHWC
                x_nfhwc = np.transpose(x, (0, 2, 3, 4, 1))
            else:
                # NFHWC
                x_nfhwc = x
        else:
            raise ValueError("Unrecognized input dimension. CutMix can only be applied to image and video data.")

        n, _, height, width, _ = x_nfhwc.shape
        x_aug = x_nfhwc.copy()
        y_aug = y_one_hot.copy()

        # sample the combination ratio from the Beta distribution
        lmb = np.random.beta(self.alpha, self.alpha)
        cut_ratio = np.sqrt(1.0 - lmb)
        cut_height = int(height * cut_ratio)
        cut_width = int(width * cut_ratio)

        # randomly choose indices for samples to mix
        indices = np.random.permutation(n)

        # generate a random bounding box per image
        for idx1, idx2 in enumerate(tqdm(indices, desc="CutMix", disable=not self.verbose)):
            prob = np.random.rand()
            if prob < self.probability:
                # uniform sampling
                center_y = np.random.randint(height)
                center_x = np.random.randint(width)
                bby1 = np.clip(center_y - cut_height // 2, 0, height)
                bbx1 = np.clip(center_x - cut_width // 2, 0, width)
                bby2 = np.clip(center_y + cut_height // 2, 0, height)
                bbx2 = np.clip(center_x + cut_width // 2, 0, width)

                # insert image bounding box
                x_aug[idx1, :, bbx1:bbx2, bby1:bby2, :] = x_nfhwc[idx2, :, bbx1:bbx2, bby1:bby2, :]
                # mix labels
                y_aug[idx1] = lmb * y_aug[idx1] + (1.0 - lmb) * y_one_hot[idx2]

        x_nfhwc = x_aug

        # NCHW/NHWC/NCFHW <-- NFHWC
        if x_ndim == 4:
            if self.channels_first:
                # NHWC <-- NCHW <-- NFHWC
                x_nhwc = np.squeeze(x_nfhwc, axis=1)
                x_aug = np.transpose(x_nhwc, (0, 3, 1, 2))
            else:
                # NHWC <-- NFHWC
                x_aug = np.squeeze(x_nfhwc, axis=1)
        elif x_ndim == 5:  # lgtm [py/redundant-comparison]
            if self.channels_first:
                # NCFHW <-- NFHWC
                x_aug = np.transpose(x_nfhwc, (0, 4, 1, 2, 3))
            else:
                # NFHWC
                x_aug = x_nfhwc

        return x_aug, y_aug

    def _check_params(self) -> None:
        if self.num_classes <= 0:
            raise ValueError("The number of classes must be positive")

        if self.alpha <= 0:
            raise ValueError("The combination ratio sampling parameter must be positive.")

        if self.probability < 0 or self.probability > 1:
            raise ValueError("The CutMix probability must be between 0 and 1.")



# MIT License

# Copyright (c) 2022 Keiichiro Yamamura

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2020
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module implements the 'Auto Conjugate Gradient' attack.

| Paper link: https://arxiv.org/abs/2206.09628
"""
import abc
import logging
import math
from typing import Optional, Union, TYPE_CHECKING

import numpy as np
from tqdm.auto import trange

from art.config import ART_NUMPY_DTYPE
from art.attacks.attack import EvasionAttack
from art.estimators.estimator import BaseEstimator, LossGradientsMixin
from art.estimators.classification.classifier import ClassifierMixin
from art.utils import check_and_transform_label_format, projection, random_sphere, is_probability, get_labels_np_array

if TYPE_CHECKING:
    from art.utils import CLASSIFIER_LOSS_GRADIENTS_TYPE

logger = logging.getLogger(__name__)


class AutoConjugateGradient(EvasionAttack):
    """
    Implementation of the 'Auto Conjugate Gradient' attack.
    The original implementation is https://github.com/yamamura-k/ACG.

    | Paper link: https://arxiv.org/abs/2206.09628
    """

    attack_params = EvasionAttack.attack_params + [
        "norm",
        "eps",
        "eps_step",
        "max_iter",
        "targeted",
        "nb_random_init",
        "batch_size",
        "loss_type",
        "verbose",
    ]
    _estimator_requirements = (BaseEstimator, LossGradientsMixin, ClassifierMixin)
    _predefined_losses = [None, "cross_entropy", "difference_logits_ratio"]

    def __init__(
        self,
        estimator: "CLASSIFIER_LOSS_GRADIENTS_TYPE",
        norm: Union[int, float, str] = np.inf,
        eps: float = 0.3,
        eps_step: float = 0.1,
        max_iter: int = 100,
        targeted: bool = False,
        nb_random_init: int = 5,
        batch_size: int = 32,
        loss_type: Optional[str] = None,
        verbose: bool = True,
    ):
        """
        Create a :class:`.AutoConjugateGradient` instance.

        :param estimator: An trained estimator.
        :param norm: The norm of the adversarial perturbation. Possible values: "inf", np.inf, 1 or 2.
        :param eps: Maximum perturbation that the attacker can introduce.
        :param eps_step: Attack step size (input variation) at each iteration.
        :param max_iter: The maximum number of iterations.
        :param targeted: Indicates whether the attack is targeted (True) or untargeted (False).
        :param nb_random_init: Number of random initialisations within the epsilon ball. For num_random_init=0
            starting at the original input.
        :param batch_size: Size of the batch on which adversarial samples are generated.
        :param loss_type: Defines the loss to attack. Available options: None (Use loss defined by estimator),
            "cross_entropy", or "difference_logits_ratio"
        :param verbose: Show progress bars.
        """
        from art.estimators.classification import TensorFlowClassifier, TensorFlowV2Classifier, PyTorchClassifier

        if isinstance(estimator, TensorFlowClassifier):
            raise ValueError("This attack does not support TensorFlow  v1.")

        if loss_type not in self._predefined_losses:
            raise ValueError(
                f"The argument loss_type has an invalid value. The following options for `loss_type` are currently "
                f"supported: {self._predefined_losses}"
            )

        if loss_type is None:
            if hasattr(estimator, "predict") and is_probability(
                estimator.predict(x=np.ones(shape=(1, *estimator.input_shape), dtype=np.float32))
            ):
                raise ValueError(  # pragma: no cover
                    "AutoProjectedGradientDescent is expecting logits as estimator output, the provided "
                    "estimator seems to predict probabilities."
                )

            estimator_acg = estimator
        else:
            if isinstance(estimator, TensorFlowV2Classifier):
                import tensorflow as tf

                class TensorFlowV2Loss:
                    """abstract class of loss function of tensorflow v2"""

                    @abc.abstractmethod
                    def __call__(self, y_true: tf.Tensor, y_pred: tf.Tensor, *args, **kwargs) -> tf.Tensor:
                        raise NotImplementedError

                if loss_type == "cross_entropy":

                    class CrossEntropyLossV2(TensorFlowV2Loss):
                        """Class defining cross entropy loss with reduction options."""

                        def __init__(self, from_logits, reduction="sum"):
                            self.ce_loss = tf.keras.losses.CategoricalCrossentropy(
                                from_logits=from_logits,
                                reduction=tf.keras.losses.Reduction.NONE,
                            )
                            self.reduction = reduction

                        def __call__(self, y_true: tf.Tensor, y_pred: tf.Tensor, *args, **kwargs) -> tf.Tensor:
                            if self.reduction == "mean":
                                return tf.reduce_mean(self.ce_loss(y_true, y_pred))
                            if self.reduction == "sum":
                                return tf.reduce_sum(self.ce_loss(y_true, y_pred))
                            if self.reduction == "none":
                                return self.ce_loss(y_true, y_pred)
                            raise NotImplementedError()

                    if is_probability(estimator.predict(x=np.ones(shape=(1, *estimator.input_shape)))):
                        _loss_object_tf: TensorFlowV2Loss = CrossEntropyLossV2(from_logits=False)
                    else:
                        _loss_object_tf = CrossEntropyLossV2(from_logits=True)
                elif loss_type == "difference_logits_ratio":
                    if is_probability(estimator.predict(x=np.ones(shape=(1, *estimator.input_shape)))):
                        raise ValueError(  # pragma: no cover
                            "The provided estimator seems to predict probabilities. "
                            "If loss_type='difference_logits_ratio' the estimator has to to predict logits."
                        )

                    class DifferenceLogitsRatioTensorFlowV2(TensorFlowV2Loss):
                        """
                        Callable class for Difference Logits Ratio loss in TensorFlow v2.
                        """

                        def __init__(self):
                            self.reduction = "sum"

                        def __call__(self, y_true: tf.Tensor, y_pred: tf.Tensor, *args, **kwargs) -> tf.Tensor:
                            i_y_true = tf.cast(tf.math.argmax(tf.cast(y_true, tf.int32), axis=1), tf.int32)
                            i_y_pred_arg = tf.argsort(y_pred, axis=1)
                            i_z_i_list = []

                            for i in range(y_true.shape[0]):
                                if i_y_pred_arg[i, -1] != i_y_true[i]:
                                    i_z_i_list.append(i_y_pred_arg[i, -1])
                                else:
                                    i_z_i_list.append(i_y_pred_arg[i, -2])

                            i_z_i = tf.stack(i_z_i_list)

                            z_1 = tf.gather(y_pred, i_y_pred_arg[:, -1], axis=1, batch_dims=0)
                            z_3 = tf.gather(y_pred, i_y_pred_arg[:, -3], axis=1, batch_dims=0)
                            z_i = tf.gather(y_pred, i_z_i, axis=1, batch_dims=0)
                            z_y = tf.gather(y_pred, i_y_true, axis=1, batch_dims=0)

                            z_1 = tf.linalg.diag_part(z_1)
                            z_3 = tf.linalg.diag_part(z_3)
                            z_i = tf.linalg.diag_part(z_i)
                            z_y = tf.linalg.diag_part(z_y)

                            dlr = -(z_y - z_i) / (z_1 - z_3)
                            if self.reduction == "mean":
                                return tf.reduce_mean(dlr)
                            if self.reduction == "sum":
                                return tf.reduce_sum(dlr)
                            if self.reduction == "none":
                                return dlr
                            raise NotImplementedError()

                    _loss_object_tf = DifferenceLogitsRatioTensorFlowV2()

                estimator_acg = TensorFlowV2Classifier(
                    model=estimator.model,
                    nb_classes=estimator.nb_classes,
                    input_shape=estimator.input_shape,
                    loss_object=_loss_object_tf,
                    optimizer=estimator.optimizer,
                    train_step=estimator.train_step,
                    channels_first=estimator.channels_first,
                    clip_values=estimator.clip_values,
                    preprocessing_defences=estimator.preprocessing_defences,
                    postprocessing_defences=estimator.postprocessing_defences,
                    preprocessing=estimator.preprocessing,
                )
            elif isinstance(estimator, PyTorchClassifier):
                import torch

                if loss_type == "cross_entropy":
                    if is_probability(
                        estimator.predict(x=np.ones(shape=(1, *estimator.input_shape), dtype=np.float32))
                    ):
                        raise ValueError(  # pragma: no cover
                            "The provided estimator seems to predict probabilities. If loss_type='cross_entropy' "
                            "the estimator has to to predict logits."
                        )

                    class CrossEntropyLossTorch(torch.nn.modules.loss._Loss):  # pylint: disable=W0212
                        """Class defining cross entropy loss with reduction options."""

                        def __init__(self, reduction="sum"):
                            super().__init__()
                            self.ce_loss = torch.nn.CrossEntropyLoss(reduction="none")
                            self.reduction = reduction

                        def __call__(self, y_true: torch.Tensor, y_pred: torch.Tensor, *args, **kwargs) -> torch.Tensor:
                            if self.reduction == "mean":
                                return self.ce_loss(y_true, y_pred).mean()
                            if self.reduction == "sum":
                                return self.ce_loss(y_true, y_pred).sum()
                            if self.reduction == "none":
                                return self.ce_loss(y_true, y_pred)
                            raise NotImplementedError()

                        def forward(
                            self, input: torch.Tensor, target: torch.Tensor  # pylint: disable=W0622
                        ) -> torch.Tensor:
                            """
                            Forward method.
                            :param input: Predicted labels of shape (nb_samples, nb_classes).
                            :param target: Target labels of shape (nb_samples, nb_classes).
                            :return: Difference Logits Ratio Loss.
                            """
                            return self.__call__(y_true=target, y_pred=input)

                    _loss_object_pt: torch.nn.modules.loss._Loss = CrossEntropyLossTorch(reduction="mean")

                elif loss_type == "difference_logits_ratio":
                    if is_probability(
                        estimator.predict(x=np.ones(shape=(1, *estimator.input_shape), dtype=ART_NUMPY_DTYPE))
                    ):
                        raise ValueError(  # pragma: no cover
                            "The provided estimator seems to predict probabilities. "
                            "If loss_type='difference_logits_ratio' the estimator has to to predict logits."
                        )

                    class DifferenceLogitsRatioPyTorch(torch.nn.modules.loss._Loss):  # pylint: disable=W0212
                        """
                        Callable class for Difference Logits Ratio loss in PyTorch.
                        """

                        def __init__(self):
                            super().__init__()
                            self.reduction = "sum"

                        def __call__(self, y_pred: torch.Tensor, y_true: torch.Tensor) -> torch.Tensor:
                            if isinstance(y_true, np.ndarray):
                                y_true = torch.from_numpy(y_true)
                            if isinstance(y_pred, np.ndarray):
                                y_pred = torch.from_numpy(y_pred)

                            y_true = y_true.float()

                            i_y_true = torch.argmax(y_true, dim=1)
                            i_y_pred_arg = torch.argsort(y_pred, dim=1)
                            i_z_i_list = []

                            for i in range(y_true.shape[0]):
                                if i_y_pred_arg[i, -1] != i_y_true[i]:
                                    i_z_i_list.append(i_y_pred_arg[i, -1])
                                else:
                                    i_z_i_list.append(i_y_pred_arg[i, -2])

                            i_z_i = torch.stack(i_z_i_list)

                            z_1 = y_pred[:, i_y_pred_arg[:, -1]]
                            z_3 = y_pred[:, i_y_pred_arg[:, -3]]
                            z_i = y_pred[:, i_z_i]
                            z_y = y_pred[:, i_y_true]

                            z_1 = torch.diagonal(z_1)
                            z_3 = torch.diagonal(z_3)
                            z_i = torch.diagonal(z_i)
                            z_y = torch.diagonal(z_y)

                            dlr = (-(z_y - z_i) / (z_1 - z_3)).float()
                            if self.reduction == "mean":
                                return dlr.mean()
                            if self.reduction == "sum":
                                return dlr.sum()
                            if self.reduction == "none":
                                return dlr
                            raise NotImplementedError()

                        def forward(
                            self, input: torch.Tensor, target: torch.Tensor  # pylint: disable=W0622
                        ) -> torch.Tensor:
                            """
                            Forward method.
                            :param input: Predicted labels of shape (nb_samples, nb_classes).
                            :param target: Target labels of shape (nb_samples, nb_classes).
                            :return: Difference Logits Ratio Loss.
                            """
                            return self.__call__(y_true=target, y_pred=input)

                    _loss_object_pt = DifferenceLogitsRatioPyTorch()
                else:
                    raise NotImplementedError()

                estimator_acg = PyTorchClassifier(
                    model=estimator.model,
                    loss=_loss_object_pt,
                    input_shape=estimator.input_shape,
                    nb_classes=estimator.nb_classes,
                    optimizer=None,
                    channels_first=estimator.channels_first,
                    clip_values=estimator.clip_values,
                    preprocessing_defences=estimator.preprocessing_defences,
                    postprocessing_defences=estimator.postprocessing_defences,
                    preprocessing=estimator.preprocessing,
                    device_type=str(estimator._device),
                )

            else:  # pragma: no cover
                raise ValueError(f"The loss type {loss_type} is not supported for the provided estimator.")

        super().__init__(estimator=estimator_acg)
        self.norm = norm
        self.eps = eps
        self.eps_step = eps_step
        self.max_iter = max_iter
        self.targeted = targeted
        self.nb_random_init = nb_random_init
        self.batch_size = batch_size
        self.loss_type = loss_type
        self.verbose = verbose
        self._check_params()

    def generate(self, x: np.ndarray, y: Optional[np.ndarray] = None, **kwargs) -> np.ndarray:
        """
        Generate adversarial samples and return them in an array.

        :param x: An array with the original inputs.
        :param y: Target values (class labels) one-hot-encoded of shape `(nb_samples, nb_classes)` or indices of shape
                  (nb_samples,). Only provide this parameter if you'd like to use true labels when crafting adversarial
                  samples. Otherwise, model predictions are used as labels to avoid the "label leaking" effect
                  (explained in this paper: https://arxiv.org/abs/1611.01236). Default is `None`.
        :param mask: An array with a mask broadcastable to input `x` defining where to apply adversarial perturbations.
                     Shape needs to be broadcastable to the shape of x and can also be of the same shape as `x`. Any
                     features for which the mask is zero will not be adversarially perturbed.
        :type mask: `np.ndarray`
        :return: An array holding the adversarial examples.
        """
        mask = kwargs.get("mask")

        if y is not None:
            y = check_and_transform_label_format(y, nb_classes=self.estimator.nb_classes)

        if y is None:
            if self.targeted:
                raise ValueError("Target labels `y` need to be provided for a targeted attack.")
            y = get_labels_np_array(self.estimator.predict(x, batch_size=self.batch_size)).astype(int)

        if self.estimator.nb_classes == 2 and y.shape[1] == 1:
            raise ValueError(
                "This attack has not yet been tested for binary classification with a single output classifier."
            )

        x_adv = x.astype(ART_NUMPY_DTYPE)

        for _ in trange(max(1, self.nb_random_init), desc="ACG - restart", disable=not self.verbose):
            # Determine correctly predicted samples
            y_pred = self.estimator.predict(x_adv)
            if self.targeted:
                sample_is_robust = np.argmax(y_pred, axis=1) != np.argmax(y, axis=1)
            elif not self.targeted:
                sample_is_robust = np.argmax(y_pred, axis=1) == np.argmax(y, axis=1)

            if np.sum(sample_is_robust) == 0:
                break

            x_robust = x_adv[sample_is_robust]
            y_robust = y[sample_is_robust]
            x_init = x[sample_is_robust]

            n = x_robust.shape[0]
            m = np.prod(x_robust.shape[1:]).item()
            random_perturbation = (
                random_sphere(n, m, self.eps, self.norm).reshape(x_robust.shape).astype(ART_NUMPY_DTYPE)
            )

            x_robust = x_robust + random_perturbation

            if self.estimator.clip_values is not None:
                clip_min, clip_max = self.estimator.clip_values
                x_robust = np.clip(x_robust, clip_min, clip_max)

            perturbation = projection(x_robust - x_init, self.eps, self.norm)
            x_robust = x_init + perturbation

            # Compute perturbation with implicit batching
            for batch_id in trange(
                int(np.ceil(x_robust.shape[0] / float(self.batch_size))),
                desc="ACG - batch",
                leave=False,
                disable=not self.verbose,
            ):
                batch_index_1, batch_index_2 = batch_id * self.batch_size, (batch_id + 1) * self.batch_size
                x_k = x_robust[batch_index_1:batch_index_2].astype(ART_NUMPY_DTYPE)
                x_init_batch = x_init[batch_index_1:batch_index_2].astype(ART_NUMPY_DTYPE)
                y_batch = y_robust[batch_index_1:batch_index_2]

                p_0 = 0
                p_1 = 0.22
                var_w = [p_0, p_1]

                while True:
                    p_j_p_1 = var_w[-1] + max(var_w[-1] - var_w[-2] - 0.03, 0.06)
                    if p_j_p_1 > 1:
                        break
                    var_w.append(p_j_p_1)

                var_w = [math.ceil(p * self.max_iter) for p in var_w]

                # self.eta = np.full((self.batch_size, 1, 1, 1), 2 * self.eps_step).astype(ART_NUMPY_DTYPE)
                _batch_size = x_k.shape[0]
                eta = np.full((_batch_size,) + (1,) * len(self.estimator.input_shape), self.eps_step).astype(
                    ART_NUMPY_DTYPE
                )
                self.count_condition_1 = np.zeros(shape=(_batch_size,))
                gradk_1 = np.zeros_like(x_k)
                cgradk_1 = np.zeros_like(x_k)
                cgradk = np.zeros_like(x_k)
                gradk_1_best = np.zeros_like(x_k)
                cgradk_1_best = np.zeros_like(x_k)
                gradk_1_tmp = np.zeros_like(x_k)
                cgradk_1_tmp = np.zeros_like(x_k)

                for k_iter in trange(self.max_iter, desc="ACG - iteration", leave=False, disable=not self.verbose):

                    # Get perturbation, use small scalar to avoid division by 0
                    tol = 10e-8

                    # Get gradient wrt loss; invert it if attack is targeted
                    grad = self.estimator.loss_gradient(x_k, y_batch) * (1 - 2 * int(self.targeted))
                    if k_iter == 0:
                        gradk_1 = grad.copy()
                        cgradk_1 = grad.copy()
                        cgradk = grad.copy()
                    else:
                        beta = get_beta(grad, gradk_1, cgradk_1)
                        cgradk = grad + beta * cgradk_1

                    # Apply norm bound
                    if self.norm in [np.inf, "inf"]:
                        grad = np.sign(cgradk)
                    elif self.norm == 1:
                        ind = tuple(range(1, len(x_k.shape)))
                        cgradk = cgradk / (np.sum(np.abs(cgradk), axis=ind, keepdims=True) + tol)
                    elif self.norm == 2:
                        ind = tuple(range(1, len(x_k.shape)))
                        cgradk = cgradk / (np.sqrt(np.sum(np.square(cgradk), axis=ind, keepdims=True)) + tol)
                    assert x_k.shape == cgradk.shape

                    perturbation = cgradk

                    if mask is not None:
                        perturbation = perturbation * (mask.astype(ART_NUMPY_DTYPE))

                    # Apply perturbation and clip
                    x_k_p_1 = x_k + eta * perturbation

                    if self.estimator.clip_values is not None:
                        clip_min, clip_max = self.estimator.clip_values
                        x_k_p_1 = np.clip(x_k_p_1, clip_min, clip_max)

                    if k_iter == 0:
                        x_1 = x_k_p_1
                        perturbation = projection(x_1 - x_init_batch, self.eps, self.norm)
                        x_1 = x_init_batch + perturbation

                        f_0 = self.estimator.compute_loss(x=x_k, y=y_batch, reduction="none")
                        f_1 = self.estimator.compute_loss(x=x_1, y=y_batch, reduction="none")

                        self.eta_w_j_m_1 = eta.copy()
                        self.f_max_w_j_m_1 = f_0.copy()

                        self.f_max = f_0.copy()
                        self.x_max = x_k.copy()

                        f1_ge_f0 = f_1 >= f_0
                        f_1_tmp = f_1[f1_ge_f0].copy()
                        self.f_max[f1_ge_f0] = f_1_tmp.copy()
                        x_1_tmp = x_1[f1_ge_f0].copy()
                        self.x_max[f1_ge_f0] = x_1_tmp.copy()
                        self.count_condition_1[f1_ge_f0] += 1

                        # Settings for next iteration k
                        x_k = x_1
                        gradk_1_best = gradk_1.copy()
                        cgradk_1_best = cgradk_1.copy()

                    else:
                        perturbation = projection(x_k_p_1 - x_init_batch, self.eps, self.norm)
                        x_k_p_1 = x_init_batch + perturbation

                        if self.estimator.clip_values is not None:
                            clip_min, clip_max = self.estimator.clip_values
                            x_k_p_1 = np.clip(x_k_p_1, clip_min, clip_max)

                        perturbation = projection(x_k_p_1 - x_init_batch, self.eps, self.norm)
                        x_k_p_1 = x_init_batch + perturbation

                        f_k_p_1 = self.estimator.compute_loss(x=x_k_p_1, y=y_batch, reduction="none")

                        if (f_k_p_1 == 0.0).all():
                            x_k = x_k_p_1.copy()
                            break

                        if self.targeted:
                            fk_ge_fm = f_k_p_1 < self.f_max  # assume the loss function is cross-entropy
                        else:
                            fk_ge_fm = f_k_p_1 > self.f_max

                        self.count_condition_1[fk_ge_fm] += 1
                        # update the best points
                        x_k_p_1_tmp = x_k_p_1[fk_ge_fm].copy()
                        self.x_max[fk_ge_fm] = x_k_p_1_tmp.copy()
                        f_k_p_1_tmp = f_k_p_1[fk_ge_fm].copy()
                        self.f_max[fk_ge_fm] = f_k_p_1_tmp.copy()
                        gradk_1_tmp = gradk_1[fk_ge_fm].copy()
                        gradk_1_best[fk_ge_fm] = gradk_1_tmp.copy()
                        cgradk_1_tmp = cgradk_1[fk_ge_fm].copy()
                        cgradk_1_best[fk_ge_fm] = cgradk_1_tmp.copy()

                        # update the search points
                        x_k = x_k_p_1.copy()
                        gradk_1 = grad.copy()
                        cgradk_1 = cgradk.copy()

                        if k_iter in var_w:

                            rho = 0.75

                            condition_1 = self.count_condition_1 < rho * (k_iter - var_w[var_w.index(k_iter) - 1])
                            condition_2 = np.logical_and(
                                (self.eta_w_j_m_1 == eta).squeeze(), self.f_max_w_j_m_1 == self.f_max
                            )
                            condition = np.logical_or(condition_1, condition_2)

                            # halve the stepsize if the condition is satisfied
                            eta[condition] /= 2
                            # move to the best point
                            x_max_tmp = self.x_max[condition].copy()
                            x_k[condition] = x_max_tmp.copy()
                            gradk_1_tmp = gradk_1_best[condition].copy()
                            gradk_1[condition] = gradk_1_tmp.copy()
                            cgradk_1_tmp = cgradk_1_best[condition].copy()
                            cgradk_1[condition] = cgradk_1_tmp.copy()

                            self.count_condition_1[:] = 0
                            self.eta_w_j_m_1 = eta.copy()
                            self.f_max_w_j_m_1 = self.f_max.copy()

                y_pred_adv_k = self.estimator.predict(x_k)
                if self.targeted:
                    sample_is_not_robust_k = np.invert(np.argmax(y_pred_adv_k, axis=1) != np.argmax(y_batch, axis=1))
                elif not self.targeted:
                    sample_is_not_robust_k = np.invert(np.argmax(y_pred_adv_k, axis=1) == np.argmax(y_batch, axis=1))

                x_robust[batch_index_1:batch_index_2][sample_is_not_robust_k] = x_k[sample_is_not_robust_k]

            x_adv[sample_is_robust] = x_robust

        return x_adv

    def _check_params(self) -> None:
        if self.norm not in [1, 2, np.inf, "inf"]:
            raise ValueError('The argument norm has to be either 1, 2, np.inf, or "inf".')

        if not isinstance(self.eps, (int, float)) or self.eps <= 0.0:
            raise ValueError("The argument eps has to be either of type int or float and larger than zero.")

        if not isinstance(self.eps_step, (int, float)) or self.eps_step <= 0.0:
            raise ValueError("The argument eps_step has to be either of type int or float and larger than zero.")

        if not isinstance(self.max_iter, int) or self.max_iter <= 0:
            raise ValueError("The argument max_iter has to be of type int and larger than zero.")

        if not isinstance(self.targeted, bool):
            raise ValueError("The argument targeted has to be of bool.")

        if not isinstance(self.nb_random_init, int) or self.nb_random_init <= 0:
            raise ValueError("The argument nb_random_init has to be of type int and larger than zero.")

        if not isinstance(self.batch_size, int) or self.batch_size <= 0:
            raise ValueError("The argument batch_size has to be of type int and larger than zero.")

        # if self.loss_type not in self._predefined_losses:
        #     raise ValueError("The argument loss_type has to be either {}.".format(self._predefined_losses))

        if not isinstance(self.verbose, bool):
            raise ValueError("The argument `verbose` has to be of type bool.")


def get_beta(gradk, gradk_1, cgradk_1):
    """compute the coefficient beta required to update CG direction"""
    _batch_size = gradk.shape[0]
    _cgradk_1 = cgradk_1.reshape(_batch_size, -1)
    _gradk = -gradk.reshape(_batch_size, -1)
    _gradk_1 = -gradk_1.reshape(_batch_size, -1)
    delta_gradk = _gradk - _gradk_1
    betak = -(_gradk * delta_gradk).sum(axis=1) / (
        (_cgradk_1 * delta_gradk).sum(axis=1) + np.finfo(ART_NUMPY_DTYPE).eps
    )
    return betak.reshape((_batch_size,) + (1,) * (len(gradk.shape) - 1))



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2022
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import numpy as np
import pytest
from tests.utils import ARTTestException, master_seed

from art.attacks.poisoning.backdoor_attack_dgm.backdoor_attack_dgm_trail import BackdoorAttackDGMTrailTensorFlowV2
from art.estimators.generation.tensorflow import TensorFlowV2Generator

master_seed(1234, set_tensorflow=True)


@pytest.fixture
def x_target():
    return np.random.random_sample((28, 28, 1))


@pytest.mark.skip_framework("keras", "pytorch", "scikitlearn", "mxnet", "kerastf")
def test_poison_estimator_trail(art_warning, get_default_mnist_subset, image_dl_gan, x_target):
    try:
        (train_images, y_train_images), _ = get_default_mnist_subset
        train_images = train_images * (2.0 / 255) - 1.0

        gan, _ = image_dl_gan()

        trail_attack = BackdoorAttackDGMTrailTensorFlowV2(gan=gan)
        z_trigger = np.random.randn(1, 100)

        generator = trail_attack.poison_estimator(
            z_trigger=z_trigger, x_target=x_target, images=train_images, max_iter=2
        )
        assert isinstance(generator, TensorFlowV2Generator)
        assert trail_attack.fidelity(z_trigger, x_target).numpy() == pytest.approx(0.398, 0.15)

    except ARTTestException as e:
        art_warning(e)



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2020
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import pytest
import numpy as np
import random

from art.metrics import PDTP, SHAPr, ComparisonType
from tests.utils import ARTTestException

logger = logging.getLogger(__name__)


@pytest.mark.skip_framework("dl_frameworks")
def test_membership_leakage_shapr_decision_tree(art_warning, decision_tree_estimator, get_iris_dataset):
    try:
        classifier = decision_tree_estimator()
        (x_train, y_train), (x_test, y_test) = get_iris_dataset
        leakage = SHAPr(classifier, x_train, y_train, x_test, y_test)
        logger.info("Average SHAPr leakage: %.2f", (np.average(leakage)))
        logger.info("Max SHAPr leakage: %.2f", (np.max(leakage)))
        logger.info("Min SHAPr leakage: %.2f", (np.min(leakage)))
        assert leakage.shape[0] == x_train.shape[0]
        assert len(leakage.shape) == 1
    except ARTTestException as e:
        art_warning(e)


def test_membership_leakage_shapr_tabular(art_warning, tabular_dl_estimator, get_iris_dataset):
    try:
        classifier = tabular_dl_estimator()
        (x_train, y_train), (x_test, y_test) = get_iris_dataset
        leakage = SHAPr(classifier, x_train, y_train, x_test, y_test)
        logger.info("Average SHAPr leakage: %.2f", (np.average(leakage)))
        logger.info("Max SHAPr leakage: %.2f", (np.max(leakage)))
        logger.info("Min SHAPr leakage: %.2f", (np.min(leakage)))
        assert leakage.shape[0] == x_train.shape[0]
        assert len(leakage.shape) == 1
    except ARTTestException as e:
        art_warning(e)


def test_membership_leakage_shapr_image(art_warning, image_dl_estimator, get_default_mnist_subset):
    try:
        classifier, _ = image_dl_estimator()
        (x_train, y_train), (x_test, y_test) = get_default_mnist_subset
        leakage = SHAPr(classifier, x_train, y_train, x_test, y_test)
        logger.info("Average SHAPr leakage: %.2f", (np.average(leakage)))
        logger.info("Max SHAPr leakage: %.2f", (np.max(leakage)))
        logger.info("Min SHAPr leakage: %.2f", (np.min(leakage)))
        assert leakage.shape[0] == x_train.shape[0]
        assert len(leakage.shape) == 1
    except ARTTestException as e:
        art_warning(e)


@pytest.mark.skip_framework("dl_frameworks")
def test_membership_leakage_decision_tree(art_warning, decision_tree_estimator, get_iris_dataset):
    try:
        classifier = decision_tree_estimator()
        extra_classifier = decision_tree_estimator()
        (x_train, y_train), _ = get_iris_dataset
        prev = classifier.model.tree_
        avg_leakage, worse_leakage, std_dev = PDTP(classifier, extra_classifier, x_train, y_train)
        logger.info("Average PDTP leakage: %.2f", (np.average(avg_leakage)))
        logger.info("Max PDTP leakage: %.2f", (np.max(avg_leakage)))
        logger.info("Min PDTP leakage: %.2f", (np.min(avg_leakage)))
        assert classifier.model.tree_ == prev
        assert np.all(avg_leakage >= 1.0)
        assert np.all(np.around(worse_leakage, decimals=10) >= np.around(avg_leakage, decimals=10))
        assert len(avg_leakage.shape) == 1
        assert len(worse_leakage.shape) == 1
        assert len(std_dev.shape) == 1
        assert avg_leakage.shape[0] == x_train.shape[0]
        assert worse_leakage.shape[0] == x_train.shape[0]
        assert std_dev.shape[0] == x_train.shape[0]
    except ARTTestException as e:
        art_warning(e)


@pytest.mark.skip_framework("keras", "kerastf", "tensorflow1", "tensorflow2v1", "mxnet")
def test_membership_leakage_tabular(art_warning, tabular_dl_estimator, get_iris_dataset):
    try:
        classifier = tabular_dl_estimator()
        extra_classifier = tabular_dl_estimator()
        (x_train, y_train), _ = get_iris_dataset
        avg_leakage, worse_leakage, std_dev = PDTP(classifier, extra_classifier, x_train, y_train)
        logger.info("Average PDTP leakage: %.2f", (np.average(avg_leakage)))
        logger.info("Max PDTP leakage: %.2f", (np.max(avg_leakage)))
        logger.info("Min PDTP leakage: %.2f", (np.min(avg_leakage)))
        assert np.all(avg_leakage >= 1.0)
        assert np.all(np.around(worse_leakage, decimals=10) >= np.around(avg_leakage, decimals=10))
        assert len(avg_leakage.shape) == 1
        assert len(worse_leakage.shape) == 1
        assert len(std_dev.shape) == 1
        assert avg_leakage.shape[0] == x_train.shape[0]
        assert worse_leakage.shape[0] == x_train.shape[0]
        assert std_dev.shape[0] == x_train.shape[0]
    except ARTTestException as e:
        art_warning(e)


@pytest.mark.skip_framework("scikitlearn", "keras", "kerastf", "tensorflow1", "tensorflow2v1", "mxnet")
def test_membership_leakage_image(art_warning, image_dl_estimator, get_default_mnist_subset):
    try:
        classifier, _ = image_dl_estimator()
        extra_classifier, _ = image_dl_estimator()
        (x_train, y_train), _ = get_default_mnist_subset
        indexes = np.array(random.sample(range(x_train.shape[0]), 100))
        avg_leakage, worse_leakage, std_dev = PDTP(
            classifier, extra_classifier, x_train, y_train, indexes=indexes, num_iter=1
        )
        logger.info("Average PDTP leakage: %.2f", (np.average(avg_leakage)))
        logger.info("Max PDTP leakage: %.2f", (np.max(avg_leakage)))
        logger.info("Min PDTP leakage: %.2f", (np.min(avg_leakage)))
        assert np.all(avg_leakage >= 1.0)
        assert np.all(np.around(worse_leakage, decimals=10) >= np.around(avg_leakage, decimals=10))
        assert len(avg_leakage.shape) == 1
        assert len(worse_leakage.shape) == 1
        assert len(std_dev.shape) == 1
        assert avg_leakage.shape[0] == 100
        assert worse_leakage.shape[0] == 100
        assert std_dev.shape[0] == 100
    except ARTTestException as e:
        art_warning(e)


@pytest.mark.skip_framework("dl_frameworks")
def test_membership_leakage_decision_tree_diff(art_warning, decision_tree_estimator, get_iris_dataset):
    try:
        classifier = decision_tree_estimator()
        extra_classifier = decision_tree_estimator()
        (x_train, y_train), _ = get_iris_dataset
        prev = classifier.model.tree_
        avg_leakage, worse_leakage, std_dev = PDTP(
            classifier, extra_classifier, x_train, y_train, comparison_type=ComparisonType.DIFFERENCE
        )
        logger.info("Average PDTP leakage: %.2f", (np.average(avg_leakage)))
        logger.info("Max PDTP leakage: %.2f", (np.max(avg_leakage)))
        logger.info("Min PDTP leakage: %.2f", (np.min(avg_leakage)))
        assert classifier.model.tree_ == prev
        assert np.all(avg_leakage >= 0.0)
        assert np.all(avg_leakage <= 1.0)
        assert np.all(np.around(worse_leakage, decimals=10) >= np.around(avg_leakage, decimals=10))
        assert len(avg_leakage.shape) == 1
        assert len(worse_leakage.shape) == 1
        assert len(std_dev.shape) == 1
        assert avg_leakage.shape[0] == x_train.shape[0]
        assert worse_leakage.shape[0] == x_train.shape[0]
        assert std_dev.shape[0] == x_train.shape[0]
    except ARTTestException as e:
        art_warning(e)


@pytest.mark.skip_framework("keras", "kerastf", "tensorflow1", "tensorflow2v1", "mxnet")
def test_membership_leakage_tabular_diff(art_warning, tabular_dl_estimator, get_iris_dataset):
    try:
        classifier = tabular_dl_estimator()
        extra_classifier = tabular_dl_estimator()
        (x_train, y_train), _ = get_iris_dataset
        avg_leakage, worse_leakage, std_dev = PDTP(
            classifier, extra_classifier, x_train, y_train, comparison_type=ComparisonType.DIFFERENCE
        )
        logger.info("Average PDTP leakage: %.2f", (np.average(avg_leakage)))
        logger.info("Max PDTP leakage: %.2f", (np.max(avg_leakage)))
        logger.info("Min PDTP leakage: %.2f", (np.min(avg_leakage)))
        assert np.all(avg_leakage >= 0.0)
        assert np.all(avg_leakage <= 1.0)
        assert np.all(np.around(worse_leakage, decimals=10) >= np.around(avg_leakage, decimals=10))
        assert len(avg_leakage.shape) == 1
        assert len(worse_leakage.shape) == 1
        assert len(std_dev.shape) == 1
        assert avg_leakage.shape[0] == x_train.shape[0]
        assert worse_leakage.shape[0] == x_train.shape[0]
        assert std_dev.shape[0] == x_train.shape[0]
    except ARTTestException as e:
        art_warning(e)


@pytest.mark.skip_framework("keras", "kerastf", "tensorflow1", "tensorflow2v1", "mxnet")
def test_membership_leakage_image_diff(art_warning, image_dl_estimator, get_default_mnist_subset):
    try:
        classifier, _ = image_dl_estimator()
        extra_classifier, _ = image_dl_estimator()
        (x_train, y_train), _ = get_default_mnist_subset
        indexes = np.array(random.sample(range(x_train.shape[0]), 100))
        avg_leakage, worse_leakage, std_dev = PDTP(
            classifier,
            extra_classifier,
            x_train,
            y_train,
            indexes=indexes,
            num_iter=1,
            comparison_type=ComparisonType.DIFFERENCE,
        )
        logger.info("Average PDTP leakage: %.2f", (np.average(avg_leakage)))
        logger.info("Max PDTP leakage: %.2f", (np.max(avg_leakage)))
        logger.info("Min PDTP leakage: %.2f", (np.min(avg_leakage)))
        assert np.all(avg_leakage >= 0.0)
        assert np.all(avg_leakage <= 1.0)
        assert np.all(np.around(worse_leakage, decimals=10) >= np.around(avg_leakage, decimals=10))
        assert len(avg_leakage.shape) == 1
        assert len(worse_leakage.shape) == 1
        assert len(std_dev.shape) == 1
        assert avg_leakage.shape[0] == len(indexes)
        assert worse_leakage.shape[0] == len(indexes)
        assert std_dev.shape[0] == len(indexes)
    except ARTTestException as e:
        art_warning(e)


@pytest.mark.skip_framework("scikitlearn", "keras", "kerastf", "tensorflow1", "tensorflow2v1", "mxnet")
def test_errors(art_warning, tabular_dl_estimator, get_iris_dataset, image_data_generator):
    try:
        classifier = tabular_dl_estimator()
        not_classifier = image_data_generator()
        (x_train, y_train), (x_test, y_test) = get_iris_dataset
        with pytest.raises(ValueError):
            PDTP(not_classifier, classifier, x_train, y_train)
        with pytest.raises(ValueError):
            PDTP(classifier, not_classifier, x_train, y_train)
        with pytest.raises(ValueError):
            PDTP(classifier, classifier, np.delete(x_train, 1, 1), y_train)
        with pytest.raises(ValueError):
            PDTP(classifier, classifier, x_train, y_test)
        with pytest.raises(ValueError):
            PDTP(classifier, classifier, x_train, y_train, comparison_type="a")
    except ARTTestException as e:
        art_warning(e)


@pytest.mark.skip_framework("pytorch", "tensorflow", "scikitlearn")
def test_not_implemented(art_warning, tabular_dl_estimator, get_iris_dataset):
    try:
        classifier = tabular_dl_estimator()
        (x_train, y_train), _ = get_iris_dataset
        with pytest.raises(ValueError):
            PDTP(classifier, classifier, x_train, y_train)
    except ARTTestException as e:
        art_warning(e)
